import unittest
from unittest.mock import patch, MagicMock

import pandas as pd

from cellmaps_vnn.annotate import VNNAnnotate


class TestVNNAnnotate(unittest.TestCase):
    """Tests for the VNNAnnotate class in the cellmaps_vnn package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        self.mock_args = MagicMock()
        self.mock_args.outdir = '/fake/output/directory'
        self.mock_args.model_predictions = ['/fake/model_predictions']
        self.mock_args.hierarchy = '/fake/hierarchy.cx2'
        self.mock_args.disease = None
        self.mock_args.upload_to_ndex = False

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_get_scores_for_disease(self):
        data = {'Term': ['Term1', 'Term2'], 'P_rho': [0.5, 0.7],
                'P_pval': [0.5, 0.7], 'C_rho': [0.5, 0.7], 'C_pval': [0.5, 0.7],
                'RLIPP': [0.5, 0.7],'Disease': ['Cancer', 'Other']}
        test_df = pd.DataFrame(data)
        annotator = VNNAnnotate(self.mock_args)
        result = annotator._get_scores_for_disease('Cancer', test_df)
        self.assertEqual(result, {'Term1': [0.5, 0.5, 0.5, 0.5, 0.5]})


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python

import argparse
import sys
import logging
import logging.config

import yaml
from cellmaps_utils import logutils
from cellmaps_utils import constants
import cellmaps_vnn
from cellmaps_vnn.annotate import VNNAnnotate
from cellmaps_vnn.exceptions import CellmapsvnnError
from cellmaps_vnn.predict import VNNPredict
from cellmaps_vnn.runner import CellmapsvnnRunner, SLURMCellmapsvnnRunner
from cellmaps_vnn.train import VNNTrain

logger = logging.getLogger(__name__)


def _parse_arguments(desc, args):
    """
    Parses command line arguments

    :param desc: description to display on command line
    :type desc: str
    :param args: command line arguments usually :py:func:`sys.argv[1:]`
    :type args: list
    :return: arguments parsed by :py:mod:`argparse`
    :rtype: :py:class:`argparse.Namespace`
    """
    parser = argparse.ArgumentParser(description=desc,
                                     formatter_class=constants.ArgParseFormatter)
    subparsers = parser.add_subparsers(dest='command', help='Command to run. Type <command> -h for more help')
    subparsers.required = True

    VNNTrain.add_subparser(subparsers)
    VNNPredict.add_subparser(subparsers)
    VNNAnnotate.add_subparser(subparsers)
    parser.add_argument('--logconf', default=None,
                        help='Path to python logging configuration file in '
                             'this format: https://docs.python.org/3/library/'
                             'logging.config.html#logging-config-fileformat '
                             'Setting this overrides -v parameter which uses '
                             ' default logger. (default None)')
    parser.add_argument('--exitcode', help='Exit code this command will return',
                        default=0, type=int)
    parser.add_argument('--skip_logging', action='store_true',
                        help='If set, output.log, error.log '
                             'files will not be created')
    parser.add_argument('--provenance',
                        help='Path to file containing provenance '
                             'information about input files in JSON format. '
                             'This is required and not including will output '
                             'and error message with example of file')
    parser.add_argument('--verbose', '-v', action='count', default=1,
                        help='Increases verbosity of logger to standard '
                             'error for log messages in this module. Messages are '
                             'output at these python logging levels '
                             '-v = WARNING, -vv = INFO, '
                             '-vvv = DEBUG, -vvvv = NOTSET (default ERROR '
                             'logging)')
    parser.add_argument('--version', action='version',
                        version=('%(prog)s ' +
                                 cellmaps_vnn.__version__))

    return parser.parse_args(args)


def main(args):
    """
    Main entry point for program

    :param args: arguments passed to command line usually :py:func:`sys.argv[1:]`
    :type args: list

    :return: return value of :py:meth:`cellmaps_vnn.runner.CellmapsvnnRunner.run`
             or ``2`` if an exception is raised
    :rtype: int
    """
    desc = """
    Version {version}

    Invokes run() method on CellmapsvnnRunner

    """.format(version=cellmaps_vnn.__version__)
    theargs = _parse_arguments(desc, args[1:])
    theargs.program = args[0]
    theargs.version = cellmaps_vnn.__version__

    if theargs.command == VNNTrain.COMMAND or theargs.command == VNNPredict.COMMAND:
        if theargs.config_file is not None:
            with open(theargs.config_file, "r") as file:
                config = yaml.safe_load(file)

                for key, value in config.items():
                    if hasattr(theargs, key):
                        setattr(theargs, key, value)

        required_args = ['cell2id', 'mutations', 'cn_deletions', 'cn_amplifications']
        if theargs.command == VNNTrain.COMMAND:
            required_args.append('gene2id')
            required_args.append('training_data')
        else:
            required_args.append('predict_data')
        for arg in required_args:
            if getattr(theargs, arg) is None:
                raise CellmapsvnnError(f"The argument --{arg} is required either in command line or config file.")

    try:
        logutils.setup_cmd_logging(theargs)

        if theargs.command == VNNTrain.COMMAND:
            cmd = VNNTrain(theargs)
        elif theargs.command == VNNPredict.COMMAND:
            cmd = VNNPredict(theargs)
        elif theargs.command == VNNAnnotate.COMMAND:
            cmd = VNNAnnotate(theargs)
            theargs.inputdir = theargs.model_predictions
        else:
            raise Exception('Invalid command: ' + str(theargs.command))

        if theargs.slurm:
            use_gpu = True if (theargs.command != VNNAnnotate.COMMAND and theargs.use_gpu) else False
            slurm_partition = 'nrnb-gpu' if (theargs.slurm_partition is None and use_gpu) else theargs.slurm_partition
            slurm_account = 'nrnb-gpu' if (theargs.slurm_account is None and use_gpu) else theargs.slurm_account

            runner = SLURMCellmapsvnnRunner(outdir=theargs.outdir,
                                            command=cmd,
                                            args=theargs,
                                            gpu=use_gpu,
                                            slurm_partition=slurm_partition,
                                            slurm_account=slurm_account,
                                            input_data_dict=theargs.__dict__
                                            )
        else:
            runner = CellmapsvnnRunner(outdir=theargs.outdir,
                                       command=cmd,
                                       inputdir=theargs.inputdir,
                                       exitcode=theargs.exitcode,
                                       skip_logging=theargs.skip_logging,
                                       input_data_dict=theargs.__dict__)

        return runner.run()
    except Exception as e:
        logger.exception('Caught exception: ' + str(e))
        return 2
    finally:
        logging.shutdown()


if __name__ == '__main__':  # pragma: no cover
    sys.exit(main(sys.argv))



import logging
import os
import shutil
from datetime import date
import getpass
import numpy as np
import pandas as pd
from cellmaps_utils.ndexupload import NDExHierarchyUploader
from cellmaps_utils import constants
from cellmaps_vnn.util import copy_and_register_gene2id_file

import cellmaps_vnn
import cellmaps_vnn.constants as vnnconstants
from ndex2.cx2 import RawCX2NetworkFactory

from cellmaps_vnn.exceptions import CellmapsvnnError

logger = logging.getLogger(__name__)


class VNNAnnotate:
    COMMAND = 'annotate'

    def __init__(self, theargs):
        """
        Constructor. Sets up the hierarchy path either directly from the arguments or by looking for
        a hierarchy.cx2 file in the first RO-Crate directory provided. If neither is found, raises an error.

        :param theargs: The arguments provided to the command line interface.
        :type theargs: argparse.Namespace
        :raises CellmapsvnnError: If no hierarchy path is specified or found.
        """
        self._theargs = theargs
        self._outdir = os.path.abspath(theargs.outdir)
        self.original_hierarchy = None
        if not os.path.exists(os.path.join(theargs.model_predictions[0], vnnconstants.RLIPP_OUTPUT_FILE)):
            theargs.model_predictions[0] = os.path.join(theargs.model_predictions[0], 'out_predict')
        if theargs.hierarchy is not None:
            self.hierarchy = theargs.hierarchy
        else:
            hierarchy_path = os.path.join(theargs.model_predictions[0], vnnconstants.HIERARCHY_FILENAME)
            if os.path.exists(hierarchy_path):
                self.hierarchy = hierarchy_path
            else:
                raise CellmapsvnnError("No hierarchy was specified or found in first ro-crate")
            original_hierarchy_path = os.path.join(theargs.model_predictions[0],
                                                   vnnconstants.ORIGINAL_HIERARCHY_FILENAME)
            if os.path.exists(original_hierarchy_path):
                self.original_hierarchy = original_hierarchy_path
        if theargs.parent_network is not None:
            self.parent_network = theargs.parent_network
        else:
            parent_network_path = os.path.join(theargs.model_predictions[0], vnnconstants.PARENT_NETWORK_NAME)
            if os.path.exists(parent_network_path):
                self.parent_network = parent_network_path
            else:
                self.parent_network = None
        if self.parent_network is not None and os.path.isfile(self.parent_network):
            self.parent_network = os.path.abspath(self.parent_network)

    @staticmethod
    def add_subparser(subparsers):
        """
        Adds a subparser for the 'annotate' command.
        """
        # TODO: modify description later
        desc = """
        Version: todo

        The 'annotate' command takes ..
        """
        parser = subparsers.add_parser(VNNAnnotate.COMMAND,
                                       help='Run prediction using a trained model',
                                       description=desc,
                                       formatter_class=constants.ArgParseFormatter)
        parser.add_argument('outdir', help='Directory to write results to')
        parser.add_argument('--model_predictions', nargs='+', required=True,
                            help='Path to one or multiple RO-Crate with the predictions and interpretations '
                                 'obtained from predict step',
                            type=str)
        parser.add_argument('--disease', help='Specify the disease or cancer type for which the annotations will be '
                                              'performed. This allows the annotation process to tailor the results '
                                              'according to the particular disease or cancer type. If not set, '
                                              'prediction scores for all diseases will be aggregated.', type=str)
        parser.add_argument('--hierarchy', help='Path to hierarchy (optional), if not set the hierarchy will be '
                                                'selected from the first RO-Crate passed in --model_predictions '
                                                'argument', type=str)
        parser.add_argument('--parent_network', help='Path to interactome (parent network) of the annotated hierarchy '
                                                     'or NDEx UUID of parent network (required if uploading '
                                                     'HCX to NDEx)', type=str)
        parser.add_argument('--ndexserver', default='ndexbio.org',
                            help='Server where annotated hierarchy will be uploaded to')
        parser.add_argument('--ndexuser',
                            help='NDEx user account. Required if uploading to NDEx.')
        parser.add_argument('--ndexpassword',
                            help='NDEx password. Enter "-" to input password interactively, or provide a file '
                                 'containing the password. Required if uploading to NDEx.')
        parser.add_argument('--visibility', action='store_true',
                            help='If set, makes Hierarchy and interactome network loaded onto '
                                 'NDEx publicly visible')
        parser.add_argument('--slurm', help='If set, slurm script for training will be generated.',
                            action='store_true')
        parser.add_argument('--slurm_partition', help='Slurm partition', type=str)
        parser.add_argument('--slurm_account', help='Slurm account', type=str)

    def _get_rlipp_out_dest_file(self):
        """
        Constructs the file path for the RLIPP output file within the specified output directory.

        :return: The file path for the RLIPP output file.
        :rtype: str
        """
        return os.path.join(self._outdir, vnnconstants.RLIPP_OUTPUT_FILE)

    def _get_hierarchy_dest_file(self):
        """
        Constructs the file path for the hierarchy output file within the specified output directory.

        :return: The file path for the hierarchy output file.
        :rtype: str
        """
        return os.path.join(self._outdir, vnnconstants.HIERARCHY_FILENAME)

    def _get_original_hierarchy_dest_file(self):
        """
        Constructs the file path for the hierarchy output file within the specified output directory.

        :return: The file path for the hierarchy output file.
        :rtype: str
        """
        return os.path.join(self._outdir, vnnconstants.ORIGINAL_HIERARCHY_FILENAME)

    def _aggregate_prediction_scores_from_models(self):
        """
        Aggregates prediction scores from multiple models' outputs by averaging them.
        The aggregated scores are then saved to the RLIPP output destination file.
        """
        data = {}

        for directory in self._theargs.model_predictions:
            if not os.path.exists(os.path.join(directory, vnnconstants.RLIPP_OUTPUT_FILE)):
                directory = os.path.join(directory, 'out_predict')
            filepath = os.path.join(directory, vnnconstants.RLIPP_OUTPUT_FILE)
            has_disease = False
            with open(filepath, 'r') as file:
                for line in file:
                    if line.startswith('Term') or not line.strip():
                        if 'Disease' in line:
                            has_disease = True
                        continue

                    parts = line.strip().split('\t')
                    if has_disease:
                        key = (parts[0], parts[-1])  # (Term, Disease)
                        values = np.array([float(v) for v in parts[1:-1]])
                    else:
                        key = (parts[0], 'unspecified')
                        values = np.array([float(v) for v in parts[1:]])

                    if key not in data:
                        data[key] = []
                    data[key].append(values)

        averaged_data = {k: np.mean(v, axis=0) for k, v in data.items()}

        with open(self._get_rlipp_out_dest_file(), 'w') as outfile:
            outfile.write("Term\tP_rho\tP_pval\tC_rho\tC_pval\tRLIPP\tDisease\n")
            for (term, disease), values in averaged_data.items():
                outfile.write(f"{term}\t" + "\t".join([f"{v:.5e}" for v in values]) + f"\t{disease}\n")

    @staticmethod
    def _aggregate_scores_from_diseases(data):
        """
        Aggregates the prediction scores for all diseases by averaging P_rho score.

        :return: A dictionary mapping each term to its averaged P_rho score across all diseases.
        :rtype: dict
        """
        aggregated_data = data.groupby('Term').agg({
            vnnconstants.PRHO_SCORE: 'mean',
            vnnconstants.P_PVAL_SCORE: 'mean',
            vnnconstants.CRHO_SCORE: 'mean',
            vnnconstants.C_PVAL_SCORE: 'mean',
            vnnconstants.RLIPP_SCORE: 'mean'
        })

        aggregated_dict = {
            term: [row[vnnconstants.PRHO_SCORE], row[vnnconstants.P_PVAL_SCORE],
                   row[vnnconstants.CRHO_SCORE], row[vnnconstants.C_PVAL_SCORE], row[vnnconstants.RLIPP_SCORE]]
            for term, row in aggregated_data.iterrows()
        }

        return aggregated_dict

    @staticmethod
    def _get_scores_for_disease(disease, data):
        """
        Retrieves prediction scores for a specific disease, returning a dictionary mapping
        each term to its P_rho score for the given disease.

        :param disease: The disease or cancer type for which scores are requested.
        :type disease: str
        :return: A dictionary with Term as keys and P_rho scores as values for the specified disease.
        :rtype: dict
        """
        filtered_data = data[data['Disease'] == disease]
        if filtered_data.empty:
            return {}

        scores = {
            term: [row[vnnconstants.PRHO_SCORE], row[vnnconstants.P_PVAL_SCORE],
                   row[vnnconstants.CRHO_SCORE], row[vnnconstants.C_PVAL_SCORE], row[vnnconstants.RLIPP_SCORE]]
            for term, row in filtered_data.set_index('Term').iterrows()
        }

        return scores

    def _upload_to_ndex_if_credentials_provided(self):
        """
        Uploads hierarchy and parent network to NDEx if credentials are provided.

        This method checks if the NDEx server, user, and password credentials are provided.
        If they are, it uploads the hierarchy and parent network to NDEx. If the parent
        network is not specified, it raises an error. If the password is specified as '-',
        it prompts the user to enter the password interactively.
        """
        if self._theargs.ndexserver and self._theargs.ndexuser and self._theargs.ndexpassword:

            if self._theargs.ndexpassword == '-':
                self._theargs.ndexpassword = getpass.getpass(prompt="Enter NDEx Password: ")

            ndex_uploader = NDExHierarchyUploader(self._theargs.ndexserver, self._theargs.ndexuser,
                                                  self._theargs.ndexpassword, self._theargs.visibility)

            if self.parent_network is None:
                logger.warning("Parent network was not specified. Hierarchy will not be in cell view.")
                cx_factory = RawCX2NetworkFactory()
                hierarchy_network = cx_factory.get_cx2network(self._get_hierarchy_dest_file())
                _, hierarchyurl = ndex_uploader._save_network(hierarchy_network)
            else:
                if os.path.isfile(self.parent_network):
                    _, _, _, hierarchyurl = ndex_uploader.upload_hierarchy_and_parent_network_from_files(
                        hierarchy_path=self._get_hierarchy_dest_file(), parent_path=self.parent_network)
                else:
                    cx_factory = RawCX2NetworkFactory()
                    hierarchy_network = cx_factory.get_cx2network(self._get_hierarchy_dest_file())
                    _, _, _, hierarchyurl = ndex_uploader.save_hierarchy_and_parent_network(hierarchy_network,
                                                                                            self.parent_network)

            print(f'Hierarchy uploaded. To view hierarchy on NDEx please paste this URL in your '
                  f'browser {hierarchyurl}. To view Hierarchy on new experimental Cytoscape on the Web, go to '
                  f'{ndex_uploader.get_cytoscape_url(hierarchyurl)}')

    @staticmethod
    def _annotate_with_score(hierarchy, original_hierarchy, node_id, score_name, score):
        hierarchy.add_node_attribute(node_id, score_name, score, datatype='double')
        if original_hierarchy is not None:
            original_hierarchy.add_node_attribute(node_id, score_name, score, datatype='double')

    def annotate(self, annotation_dict):
        """
        Annotates the hierarchy with P_rho scores from the given annotation dictionary,
        updating node attributes within the hierarchy file.

        :param annotation_dict: A dictionary mapping terms to their P_rho scores.
        :type annotation_dict: dict
        """
        factory = RawCX2NetworkFactory()
        hierarchy = factory.get_cx2network(self.hierarchy)
        original_hierarchy = None
        if self.original_hierarchy is not None:
            original_hierarchy = factory.get_cx2network(self.original_hierarchy)

        for term, score in annotation_dict.items():
            node_id = term
            if not isinstance(term, int):
                node_id = hierarchy.lookup_node_id_by_name(term)
            if node_id is not None:
                self._annotate_with_score(hierarchy, original_hierarchy, node_id, vnnconstants.PRHO_SCORE, score[0])
                self._annotate_with_score(hierarchy, original_hierarchy, node_id, vnnconstants.P_PVAL_SCORE, score[1])
                self._annotate_with_score(hierarchy, original_hierarchy, node_id, vnnconstants.CRHO_SCORE, score[2])
                self._annotate_with_score(hierarchy, original_hierarchy, node_id, vnnconstants.C_PVAL_SCORE, score[3])
                self._annotate_with_score(hierarchy, original_hierarchy, node_id, vnnconstants.RLIPP_SCORE, score[4])

        # TODO: apply style to the hierarchy
        path_to_style_network = os.path.join(os.path.dirname(cellmaps_vnn.__file__), 'nest_style.cx2')
        style_network = factory.get_cx2network(path_to_style_network)
        vis_prop = style_network.get_visual_properties()
        hierarchy.set_visual_properties(vis_prop)
        hierarchy.write_as_raw_cx2(self._get_hierarchy_dest_file())
        if original_hierarchy is not None:
            original_hierarchy.set_visual_properties(vis_prop)
            original_hierarchy.write_as_raw_cx2(self._get_original_hierarchy_dest_file())

    def run(self):
        """
        The logic for annotating hierarchy with prediction results from cellmaps_vnn. It aggregates prediction scores
        from models, optionally filters them for a specific disease, and annotates the hierarchy with these scores.
        """
        self._aggregate_prediction_scores_from_models()
        filepath = self._get_rlipp_out_dest_file()
        data = pd.read_csv(filepath, sep='\t')
        if self._theargs.disease is None:
            annotation_dict = self._aggregate_scores_from_diseases(data)
        else:
            annotation_dict = self._get_scores_for_disease(self._theargs.disease, data)
        if len(annotation_dict) == 0:
            print("No system importance scores available for annotation. Training was not sufficient. "
                  "Increase number of epochs and run train and predict again.")
            raise CellmapsvnnError("No system importance scores available for annotation. "
                                   "Please ensure valid data is provided for the hierarchy annotation.")
        self.annotate(annotation_dict)
        self._upload_to_ndex_if_credentials_provided()

    def register_outputs(self, outdir, description, keywords, provenance_utils):
        """
        Registers the output files of the annotation process with the FAIRSCAPE service for data provenance.
        This includes the annotated hierarchy and the RLIPP output files.

        :param outdir: The output directory where the files are stored.
        :type outdir: str
        :param description: A description of the files for provenance registration.
        :type description: str
        :param keywords: A list of keywords associated with the files.
        :type keywords: list
        :param provenance_utils: The utility class for provenance registration.
        :type provenance_utils: ProvenanceUtility
        :return: A list of dataset IDs assigned to the registered files.
        :rtype: list
        """
        hierarchy_id = self._register_hierarchy(outdir, description, keywords, provenance_utils)
        rlipp_id = self._register_rlipp_file(outdir, description, keywords, provenance_utils)
        return_ids = [hierarchy_id, rlipp_id]
        gene2ind_path = os.path.join(self._theargs.model_predictions[0], 'gene2ind.txt')
        if os.path.exists(gene2ind_path):
            gene2ind_id = copy_and_register_gene2id_file(gene2ind_path, outdir, description, keywords,
                                                         provenance_utils)
            return_ids.append(gene2ind_id)
        if self.original_hierarchy is not None:
            original_hierarchy_id = self._register_original_hierarchy(outdir, description, keywords, provenance_utils)
            return_ids.append(original_hierarchy_id)
        if self.parent_network is not None and os.path.isfile(self.parent_network):
            hierarchy_parent_id = self._copy_and_register_hierarchy_parent(outdir, description, keywords,
                                                                           provenance_utils)
            return_ids.append(hierarchy_parent_id)
        return return_ids

    def _register_hierarchy(self, outdir, description, keywords, provenance_utils):
        """
        Register annotated hierarchy file with the FAIRSCAPE service for data provenance.

        :param outdir: The output directory where the outputs are stored.
        :param description: Description of the file for provenance registration.
        :param keywords: List of keywords associated with the file.
        :param provenance_utils: The utility class for provenance registration.

        :return: The dataset ID assigned to the registered file.
        """
        hierarchy_out_file = self._get_original_hierarchy_dest_file()

        data_dict = {'name': os.path.basename(hierarchy_out_file) + ' Annotated hierarchy file that was used to build '
                                                                    'VNN',
                     'description': description + ' Annotated hierarchy file that was used to build VNN',
                     'keywords': keywords,
                     'data-format': 'CX2',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime('%m-%d-%Y')}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=hierarchy_out_file,
                                                       data_dict=data_dict)
        return dataset_id

    def _register_original_hierarchy(self, outdir, description, keywords, provenance_utils):
        """
        Register annotated hierarchy file with the FAIRSCAPE service for data provenance.

        :param outdir: The output directory where the outputs are stored.
        :param description: Description of the file for provenance registration.
        :param keywords: List of keywords associated with the file.
        :param provenance_utils: The utility class for provenance registration.

        :return: The dataset ID assigned to the registered file.
        """
        hierarchy_out_file = self._get_hierarchy_dest_file()

        data_dict = {'name': os.path.basename(hierarchy_out_file) + ' Annotated hierarchy file',
                     'description': description + ' Annotated hierarchy file',
                     'keywords': keywords,
                     'data-format': 'CX2',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime('%m-%d-%Y')}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=hierarchy_out_file,
                                                       data_dict=data_dict)
        return dataset_id

    def _copy_and_register_hierarchy_parent(self, outdir, description, keywords, provenance_utils):
        hierarchy_parent_out_file = os.path.join(outdir, 'hierarchy_parent.cx2')
        shutil.copy(self.parent_network, hierarchy_parent_out_file)

        data_dict = {'name': os.path.basename(hierarchy_parent_out_file) + ' Hierarchy parent network file',
                     'description': description + ' Hierarchy parent network file',
                     'keywords': keywords,
                     'data-format': 'CX2',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime('%m-%d-%Y')}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=hierarchy_parent_out_file,
                                                       data_dict=data_dict)
        return dataset_id

    def _register_rlipp_file(self, outdir, description, keywords, provenance_utils):
        """
        Registers the rlipp aggregated file with the FAIRSCAPE service for data provenance.

        :param outdir: The output directory where the outputs are stored.
        :param description: Description of the file for provenance registration.
        :param keywords: List of keywords associated with the file.
        :param provenance_utils: The utility class for provenance registration.

        :return: The dataset ID assigned to the registered file.
        """
        dest_path = self._get_rlipp_out_dest_file()
        description = description
        description += ' rlipp results file averaged with multiple models'
        keywords = keywords
        keywords.extend(['file'])
        data_dict = {'name': os.path.basename(dest_path) + ' rlipp aggregated file',
                     'description': description,
                     'keywords': keywords,
                     'data-format': 'txt',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime(provenance_utils.get_default_date_format_str())}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=dest_path,
                                                       data_dict=data_dict)
        return dataset_id



# -*- coding: utf-8 -*-

"""Top-level package for cellmaps_vnn."""

__author__ = """Cell Maps team"""
__email__ = 'tools@cm4ai.org'
__version__ = '0.1.0a10'
__repo_url__ = 'https://github.com/idekerlab/cellmaps_vnn'
__description__ = 'Cell Maps Visual Neural Network Toolkit'



import copy
import json
import os.path
import sys

import ndex2
import networkx as nx
import logging
import pandas as pd
import random as rd
import torch
from ndex2.cx2 import RawCX2NetworkFactory, CX2NetworkXFactory
import cellmaps_vnn.constants as constants

import cellmaps_vnn.util as util
from cellmaps_vnn.exceptions import CellmapsvnnError

logger = logging.getLogger(__name__)

class TrainingDataWrapper:

    def __init__(self, theargs):
        """
        Initializes the TrainingDataWrapper object with configuration and training data parameters.

        :param theargs: Arguments containing configuration and training parameters.
        """

        self.root = None
        self.digraph = None
        self.num_hiddens_genotype = theargs.genotype_hiddens
        self.lr = theargs.lr
        self.wd = theargs.wd
        self.alpha = theargs.alpha
        self.epochs = theargs.epoch
        self.batchsize = theargs.batchsize
        self.outdir = theargs.outdir
        self.modelfile = theargs.modelfile
        self.cuda = theargs.cuda
        self.zscore_method = theargs.zscore_method
        self.std = theargs.stdfile
        self.patience = theargs.patience
        self.delta = theargs.delta
        self.min_dropout_layer = theargs.min_dropout_layer
        self.dropout_fraction = theargs.dropout_fraction
        self.gene_attribute_name = theargs.gene_attribute_name

        self._hierarchy = os.path.join(theargs.inputdir, constants.HIERARCHY_FILENAME)
        self._training_data = theargs.training_data
        self.cell_id_mapping = util.load_mapping(theargs.cell2id, 'cell lines')
        self.gene_id_mapping = util.load_mapping(theargs.gene2id, 'genes')

        self.cell_features = util.load_cell_features(theargs.mutations, theargs.cn_deletions, theargs.cn_amplifications)
        self.train_feature, self.train_label, self.val_feature, self.val_label = self._prepare_train_data()
        self._load_graph(self._hierarchy)

    def _prepare_train_data(self):
        """
        Prepares the training data for model training.

        :return: Tensors for train features, train labels, validation features, and validation labels.
        :rtype: Tuple(Tensor, FloatTensor, Tensor, FloatTensor)
        """
        train_features, train_labels, val_features, val_labels = self._load_train_data()
        return (torch.Tensor(train_features), torch.FloatTensor(train_labels), torch.Tensor(val_features),
                torch.FloatTensor(val_labels))

    def _load_train_data(self):
        """
        Loads and processes the training data from a file.

        :return: Features and labels for both training and validation data.
        :rtype: Tuple(List, List, List, List)
        """
        all_df = pd.read_csv(self._training_data, sep='\t', header=None,
                             names=['cell_line', 'smiles', 'auc', 'dataset'])
        filtered_df = all_df[all_df['cell_line'].isin(self.cell_id_mapping.keys())]
        train_df, val_df = self._split_train_val_data(filtered_df)
        std_df = util.calc_std_vals(train_df, self.zscore_method)
        std_df.to_csv(self.std, sep='\t', header=False, index=False)
        train_df = util.standardize_data(train_df, std_df)
        val_df = util.standardize_data(val_df, std_df)
        train_features, train_labels = self._extract_features_labels(train_df)
        val_features, val_labels = self._extract_features_labels(val_df)
        return train_features, train_labels, val_features, val_labels

    def _split_train_val_data(self, all_df):
        """
        Splits the data into training and validation datasets.

        :param all_df: DataFrame containing the entire dataset.
        :type all_df: pandas.DataFrame

        :return: DataFrames for training and validation datasets.
        :rtype: Tuple(pandas.DataFrame, pandas.DataFrame)
        """
        train_cell_lines = list(set(all_df['cell_line']))
        val_cell_lines = self._select_validation_cell_lines(train_cell_lines)
        val_df = all_df.query('cell_line in @val_cell_lines').reset_index(drop=True)
        train_df = all_df.query('cell_line in @train_cell_lines').reset_index(drop=True)
        return train_df, val_df

    @staticmethod
    def _select_validation_cell_lines(train_cell_lines):
        """
        Selects cell lines for validation from the training cell lines.

        :param train_cell_lines: List of cell lines to choose from for validation.
        :type train_cell_lines: List

        :return: Selected cell lines for validation.
        :rtype: List
        """
        val_size = int(len(train_cell_lines) / 5)
        val_cell_lines = []
        for _ in range(val_size):
            r = rd.randint(0, len(train_cell_lines) - 1)
            val_cell_lines.append(train_cell_lines.pop(r))
        return val_cell_lines

    def _extract_features_labels(self, df):
        """
        Extracts features and labels from a data frame.

        :param df: DataFrame containing the data.
        :type df: pandas.DataFrame

        :return: Extracted features and labels.
        :rtype: Tuple(List, List)
        """
        features = []
        labels = []
        for row in df.values:
            features.append([self.cell_id_mapping[row[0]]])
            labels.append([float(row[2])])
        return features, labels

    def _load_graph(self, file_name):
        """
        Loads a graph from a file and performs initial processing.

        :param file_name: Name of the file containing the graph data.
        :type file_name: str

        :raises CellmapsvnnError: If the graph does not meet specified criteria.
        """

        try:
            digraph, cx2network = self._create_digraph(file_name)
            roots = [n for n in digraph.nodes if digraph.in_degree(n) == 0]
            ugraph = digraph.to_undirected()
            connected_sub_graph_list = list(nx.connected_components(ugraph))

            if len(roots) != 1 or len(connected_sub_graph_list) != 1:
                raise CellmapsvnnError("Graph must have exactly one root and be fully connected")

            self.digraph = self._convert_graph_to_string_nodes(digraph)
            self.root = str(roots[0])
            self._generate_term_maps(cx2network)

        except Exception as e:
            raise CellmapsvnnError(f"Error loading graph: {e}")

    @staticmethod
    def _create_digraph(file_name):
        """
        Creates a directed graph from a given file.

        :param file_name: Name of the file containing the graph data.
        :type file_name: str

        :return: A directed graph and its CX2 network representation.
        :rtype: Tuple(nx.DiGraph, CX2Network)
        """
        cx2factory = RawCX2NetworkFactory()
        nxfactory = CX2NetworkXFactory()
        cx2network = cx2factory.get_cx2network(file_name)
        digraph = nxfactory.get_graph(cx2network, nx.DiGraph())
        return digraph, cx2network

    def _generate_term_maps(self, cx2_network):
        """
        Generates term maps from a CX2 network.

        :param cx2_network: CX2 network representation of the graph.
        :type cx2_network: CX2Network
        """
        term_direct_gene_map = self._get_direct_genes(cx2_network)
        term_size_map = {}
        empty_terms = []
        pruned_hierarchy = copy.deepcopy(cx2_network)

        for term in self.digraph.nodes():
            term_gene_set = term_direct_gene_map.get(term, set())
            descendants = nx.descendants(self.digraph, term)
            for child in descendants:
                if child in term_direct_gene_map:
                    term_gene_set = term_gene_set | term_direct_gene_map[child]

            if len(term_gene_set) == 0:
                logger.warning("There is an empty term, it will not be part of the VNN.")
                empty_terms.append(term)
                pruned_hierarchy.remove_node(int(term))
                if term in term_direct_gene_map:
                    del term_direct_gene_map[term]
            else:
                term_size_map[term] = len(term_gene_set)

        if empty_terms:
            output_path = os.path.join(self.outdir, 'vnn_excluded_terms.txt')
            with open(output_path, 'w') as file:
                for term in empty_terms:
                    file.write(f'{term}\n')

        hierarchy_json = pruned_hierarchy.to_cx2()
        for item in hierarchy_json:
            if 'nodeBypasses' in item:
                item['nodeBypasses'] = []
            if 'edgeBypasses' in item:
                item['edgeBypasses'] = []

        with open(os.path.join(self.outdir, constants.HIERARCHY_FILENAME), 'w') as output_file:
            json.dump(hierarchy_json, output_file, indent=4)

        self.digraph.remove_nodes_from(empty_terms)

        self.term_size_map = term_size_map
        self.term_direct_gene_map = term_direct_gene_map

    def _get_direct_genes(self, cx2_network):
        """
        Extracts direct gene associations from a CX2 network.

        :param cx2_network: CX2 network representation of the graph.
        :type cx2_network: CX2Network

        :return: A map of terms to their directly associated genes.
        :rtype: Dict
        """
        term_direct_gene_map = {}
        child_genes_map = {}

        for edge_id, edge_data in cx2_network.get_edges().items():
            parent_node_id = edge_data['s']
            child_node_id = edge_data['t']

            if parent_node_id not in child_genes_map:
                child_genes_map[parent_node_id] = set()

            child_node_genes = self._get_genes_of_node(cx2_network, child_node_id)
            child_genes_map[parent_node_id].update(child_node_genes)

        for node_id, node_data in cx2_network.get_nodes().items():
            node_id_str = str(node_id)
            all_genes = self._get_genes_of_node(cx2_network, node_id)
            direct_genes = all_genes - child_genes_map.get(node_id, set())
            if len(direct_genes) > 0:
                term_direct_gene_map[node_id_str] = direct_genes

        return term_direct_gene_map

    def _get_genes_of_node(self, cx2_network, node_id):
        """
        Retrieves genes associated with a specific node in the CX2 network.

        :param cx2_network: CX2 network representation of the graph.
        :type cx2_network: CX2Network
        :param node_id: The ID of the node.
        :type node_id: int

        :return: A set of genes associated with the node.
        :rtype: Set
        """
        genes = set()
        node_data = cx2_network.get_node(node_id)

        if node_data and self.gene_attribute_name in node_data[ndex2.constants.ASPECT_VALUES]:
            for gene_identifier in node_data[ndex2.constants.ASPECT_VALUES][self.gene_attribute_name].split():
                if gene_identifier in self.gene_id_mapping:
                    genes.add(self.gene_id_mapping[gene_identifier])

        return genes

    @staticmethod
    def _convert_graph_to_string_nodes(original_graph):
        """
        Converts a graph with integer nodes to a graph with string nodes.

        :param original_graph: The original graph with integer nodes.
        :type original_graph: nx.Graph

        :return new_graph: A new graph with the same structure but with string nodes.
        :rtype new_graph: nx.Graph
        """
        new_graph = type(original_graph)()

        for node in original_graph.nodes():
            new_graph.add_node(str(node))

        for edge in original_graph.edges():
            new_graph.add_edge(str(edge[0]), str(edge[1]))

        return new_graph



import os

import numpy as np
import pandas as pd
import time
from scipy import stats
from multiprocessing import Pool
from joblib import Parallel, delayed
from sklearn.decomposition import PCA
from sklearn.linear_model import RidgeCV
import logging
from cellmaps_vnn.exceptions import CellmapsvnnError

logger = logging.getLogger(__name__)


class RLIPPCalculator:
    """
    A calculator for Relative Importance of Predictor Performance (RLIPP) scores.

    Parameters:
    outdir (str): Output directory for the RLIPP scores and gene correlations.
    hierarchy (CX2Network): A hierarchy in HCX format.
    test_data (str):
    predicted_data (str): Path to the file containing predicted values.
    gene2idfile (str): Path to the file mapping genes to IDs.
    cell2idfile (str): Path to the file mapping cells to IDs.
    hidden_dir (str): Directory containing hidden layer outputs.
    rlipp_file (str): Path of the output file where results of rlipp algorithm will be saved
    gene_rho_file (str): Path of the output file where gene rho scores will be saved
    cpu_count (int): No of available cores
    num_hiddens_genotype (int): Mapping for the number of neurons in each term in genotype parts
    drug_count (int): No of top performing drugs
    """

    def __init__(self, hierarchy, test_data, predicted_data, gene2idfile, cell2idfile, hidden_dir,
                 rlipp_file, gene_rho_file, cpu_count, num_hiddens_genotype, drug_count, excluded_terms=[]):
        self._hierarchy = hierarchy
        all_terms = list(hierarchy.get_nodes().keys())
        self.terms = [term for term in all_terms if term not in list(excluded_terms)]

        try:
            test_df = pd.read_csv(test_data, sep='\t', header=None, names=['C', 'D', 'AUC', 'DS'])
        except Exception as e:
            raise CellmapsvnnError(f"Failed to read test data from {test_data}: {e}")

        try:
            self.predicted_vals = np.loadtxt(predicted_data)
        except Exception as e:
            raise CellmapsvnnError(f"Failed to load predicted values from {predicted_data}: {e}")

        try:
            self.genes = pd.read_csv(gene2idfile, sep='\t', header=None, names=['I', 'G'])['G']
        except Exception as e:
            raise CellmapsvnnError(f"Failed to read gene ID file from {gene2idfile}: {e}")

        try:
            cell_index = pd.read_csv(cell2idfile, sep="\t", header=None, names=['I', 'C'])
        except Exception as e:
            raise CellmapsvnnError(f"Failed to read cell ID file from {cell2idfile}: {e}")

        self.test_df = test_df[test_df['C'].isin(list(cell_index['C']))].reset_index(drop=True)

        self.hidden_dir = hidden_dir
        self.rlipp_file = rlipp_file
        self.gene_rho_file = gene_rho_file
        self.cpu_count = cpu_count
        self.num_hiddens_genotype = num_hiddens_genotype
        self.drug_count = drug_count
        self.drugs = list(set(self.test_df['D']))
        if self.drug_count == 0:
            self.drug_count = len(self.drugs)

    def create_drug_pos_map(self):
        """
        Creates a mapping from drugs to their positions in the test data file.

        :return: A dictionary where keys are drugs and values are lists of positions in the test data.
        :rtype: dict
        """
        drug_pos_map = {}
        for i, row in self.test_df.iterrows():
            if row['D'] not in drug_pos_map:
                drug_pos_map[row['D']] = []
            drug_pos_map[row['D']].append(i)
        return drug_pos_map

    def create_drug_corr_map_sorted(self, drug_pos_map):
        """
        Creates a sorted mapping of drugs to their Spearman correlation values.

        :param drug_pos_map: A dictionary mapping drugs to their positions in the test data.
        :type drug_pos_map: dict

        :return: A dictionary of drugs sorted by their Spearman correlation values in descending order.
        :rtype: dict
        """
        test_auc = np.array(self.test_df['AUC'])
        drug_corr_map = {}
        for drug, positions in drug_pos_map.items():
            if not positions:
                drug_corr_map[drug] = 0.0
                continue

            try:
                test_vals = np.take(test_auc, positions)
                pred_vals = np.take(self.predicted_vals, positions)
                drug_corr_map[drug] = stats.spearmanr(test_vals, pred_vals)[0]
            except IndexError:
                drug_corr_map[drug] = 0.0

        return {drug: corr for drug, corr in sorted(drug_corr_map.items(), key=lambda item: item[1], reverse=True)}

    def load_feature(self, element, size):
        """
        Loads hidden features for a given element.

        :param element: The element (term or gene) whose features are to be loaded.
        :type element: str
        :param size: The number of columns (features) to load.
        :type size: int

        :return: A numpy array of the hidden features for the given element.
        :rtype: numpy.ndarray
        """
        file_name = self.hidden_dir + "/" + str(element) + '.hidden'
        return np.loadtxt(file_name, usecols=range(size))

    def load_term_features(self, term):
        """
        Loads hidden features for a given term.

        :param term: The term whose features are to be loaded.
        :type term: str

        :return: A numpy array of the hidden features for the given term.
        :rtype: numpy.ndarray
        """
        return self.load_feature(term, self.num_hiddens_genotype)

    def load_gene_features(self, gene):
        """
        Loads hidden features for a given gene.

        :param gene: The gene whose features are to be loaded.
        :type gene: str

        :return: A numpy array of the hidden features for the given gene.
        :rtype: numpy.ndarray
        """
        return self.load_feature(gene, 1)

    def create_child_feature_map(self, feature_map, term):
        """
        Creates a map of child features for a given term.

        :param feature_map: A dictionary mapping terms/genes to their features.
        :type feature_map: dict
        :param term: The term for which child features are to be created.
        :type term: str

        :return: A list of child features for the given term.
        :rtype: list
        """
        child_features = [term]
        child_features.extend(
            feature_map[edge_data['t']] for _, edge_data in self._hierarchy.get_edges().items() if
            edge_data['s'] == term)
        return child_features

    def load_all_features(self):
        """
        Loads hidden features for all terms and genes.

        :return: A tuple containing two dictionaries, one mapping terms/genes to their features and the other mapping
                 terms to their child features.
        :rtype: (dict, dict)
        """
        feature_map = {}

        # Load term and gene features
        with Pool(self.cpu_count) as p:
            term_features = p.map(self.load_term_features, self.terms)
            gene_features = p.map(self.load_gene_features, self.genes)

        # Merge results into the feature map
        for i, term in enumerate(self.terms):
            feature_map[term] = term_features[i]
        for i, gene in enumerate(self.genes):
            feature_map[gene] = gene_features[i]

        # Build child feature map
        child_feature_map = {}
        for term in self.terms:
            children = [edge_data['t'] for edge_id, edge_data in self._hierarchy.get_edges().items() if
                        edge_data['s'] == term]
            child_feature_map[term] = [feature_map[child] for child in children if child in feature_map]

        return feature_map, child_feature_map

    @staticmethod
    def get_child_features(term_child_features, position_map):
        """
        Gets a matrix of hidden features for a given term's children.

        :param term_child_features: A list of features for the children of a term.
        :type term_child_features: list
        :param position_map: A list of positions for which features are to be extracted.
        :type position_map: list

        :return: A matrix of hidden features for the children of the given term.
        :rtype: numpy.ndarray
        """
        child_features = []
        for f in term_child_features:
            child_features.append(np.take(f, position_map, axis=0))
        return np.column_stack([f for f in child_features])

    def exec_lm(self, X, y):
        """
        Executes 5-fold cross-validated Ridge regression for a given hidden features matrix
        and returns the Spearman correlation value of the predicted output.

        :param X: The input matrix for regression.
        :type X: numpy.ndarray
        :param y: The target variable.
        :type y: numpy.ndarray

        :return: A tuple containing the Spearman correlation coefficient and p-value.
        :rtype: (float, float)
        """
        pca = PCA(n_components=self.num_hiddens_genotype)
        X_pca = pca.fit_transform(X)

        regr = RidgeCV(cv=5)
        regr.fit(X_pca, y)
        y_pred = regr.predict(X_pca)
        return stats.spearmanr(y_pred, y)

    def calc_term_rlipp(self, term_features, term_child_features, position_map, term, drug):
        """
        Calculates the RLIPP score for a given term and drug.

        :param term_features: The features for the parent term.
        :type term_features: numpy.ndarray
        :param term_child_features: The features for the children of the term.
        :type term_child_features: list
        :param position_map: A list of positions for which RLIPP is to be calculated.
        :type position_map: list
        :param term: The term for which RLIPP is calculated.
        :type term: str
        :param drug: The drug for which RLIPP is calculated.
        :type drug: str

        :return: A formatted string containing the term, Spearman correlations, p-values, and RLIPP score.
        :rtype: str
        """
        if not term_child_features:
            return ''
        X_parent = np.take(term_features, position_map, axis=0)
        X_child = self.get_child_features(term_child_features, position_map)
        y = np.take(self.predicted_vals, position_map)
        p_rho, p_pval = self.exec_lm(X_parent, y)
        c_rho, c_pval = self.exec_lm(X_child, y)

        message = (f"The model was not sufficiently trained - the system importance scores cannot be calculated "
                   f"correctly. ")

        if c_rho == 0:
            reason = f"Reason: Division by zero error: c_rho is zero for term '{term}' and drug '{drug}'"
            raise CellmapsvnnError(message + reason)

        rlipp = p_rho / c_rho

        if np.isnan(rlipp) or np.isinf(rlipp):
            reason = f"Reason: Invalid RLIPP value: {rlipp} for term '{term}' and drug '{drug}'"
            raise CellmapsvnnError(message + reason)

        rlipp = p_rho / c_rho
        result = '{}\t{:.3e}\t{:.3e}\t{:.3e}\t{:.3e}\t{:.3e}\n'.format(term, p_rho, p_pval, c_rho, c_pval, rlipp)
        return result

    def calc_gene_rho(self, gene_features, position_map, gene, drug):
        """
        Calculates Spearman correlation between gene embeddings and predicted AUC.

        :param gene_features: The features for the gene.
        :type gene_features: numpy.ndarray
        :param position_map: A list of positions for which correlation is to be calculated.
        :type position_map: list
        :param gene: The gene for which correlation is calculated.
        :type gene: str
        :param drug: The drug for which correlation is calculated.
        :type drug: str

        :return: A formatted string containing the gene, Spearman correlation, and p-value.
        :rtype: str
        """
        pred = np.take(self.predicted_vals, position_map)
        gene_embeddings = np.take(gene_features, position_map)
        rho, p_val = stats.spearmanr(pred, gene_embeddings)
        result = '{}\t{:.3e}\t{:.3e}\n'.format(gene, rho, p_val)
        return result

    def calc_scores(self):
        """
        Calculates RLIPP scores for top n drugs (n = drug_count),
        and prints the result in "Drug Term P_rho C_rho RLIPP" format.

        This method runs the calculation in parallel for efficiency.
        """
        logger.info('Starting prediction process')
        print('Starting score calculation')
        drug_pos_map = self.create_drug_pos_map()
        sorted_drugs = list(self.create_drug_corr_map_sorted(drug_pos_map).keys())[0:self.drug_count]

        start = time.time()
        feature_map, child_feature_map = self.load_all_features()
        time_passed = time.time() - start
        logger.info('Time taken to load features: {:.4f}'.format(time_passed))

        with open(self.rlipp_file, "w") as rlipp_file, open(self.gene_rho_file, "w") as gene_rho_file:
            rlipp_file.write('Term\tP_rho\tP_pval\tC_rho\tC_pval\tRLIPP\n')
            gene_rho_file.write('Gene\tRho\tP_val\n')

            with Parallel(backend="multiprocessing", n_jobs=self.cpu_count) as parallel:
                for i, drug in enumerate(sorted_drugs):
                    try:
                        start = time.time()

                        # Parallel computation of RLIPP and gene correlation results
                        rlipp_results = parallel(
                            delayed(self.calc_term_rlipp)(feature_map[term], child_feature_map[term],
                                                          drug_pos_map[drug], term,
                                                          drug) for term in self.terms)
                        gene_rho_results = parallel(
                            delayed(self.calc_gene_rho)(feature_map[gene], drug_pos_map[drug], gene, drug) for gene in
                            self.genes)

                        # After collecting all results, write them to files
                        for result in rlipp_results:
                            rlipp_file.write(result)
                        for result in gene_rho_results:
                            gene_rho_file.write(result)

                        time_passed = time.time() - start
                        logger.info('Drug {} completed in {:.4f} seconds'.format((i + 1), time_passed))
                    except CellmapsvnnError as e:
                        print(e)
                        raise CellmapsvnnError(e)
                    except Exception as e:
                        logger.warning(f"Error during processing for drug {drug}: {e}")
                        print(f"Error during processing for drug {drug}: {e}")



import os
import logging
import shutil
from datetime import date

from cellmaps_vnn.util import copy_and_register_gene2id_file
from tqdm import tqdm
import numpy as np
import pandas as pd
import torch
import torch.utils.data as du
from torch.autograd import Variable
from cellmaps_utils import constants
import cellmaps_vnn.constants as vnnconstants
from ndex2.cx2 import RawCX2NetworkFactory

import cellmaps_vnn
from cellmaps_vnn import util
from cellmaps_vnn.exceptions import CellmapsvnnError
from cellmaps_vnn.rlipp_calculator import RLIPPCalculator

logger = logging.getLogger(__name__)


class VNNPredict:
    COMMAND = 'predict'

    def __init__(self, theargs):
        """
        Constructor for predicting with a trained model.
        """
        if not os.path.exists(os.path.join(theargs.inputdir, 'model_final.pt')):
            theargs.inputdir = os.path.join(theargs.inputdir, 'out_train')
        self._theargs = theargs
        self._outdir = os.path.abspath(theargs.outdir)
        self._number_feature_grads = 0
        self.use_cuda = torch.cuda.is_available() and self._theargs.cuda is not None

        self.excluded_terms = []
        excluded_terms_path = os.path.join(theargs.inputdir, 'vnn_excluded_terms.txt')
        if os.path.exists(excluded_terms_path):
            with open(excluded_terms_path, 'r') as file:
                self.excluded_terms = set(int(line.strip()) for line in file if line.strip().isdigit())

    @staticmethod
    def add_subparser(subparsers):
        """
        Adds a subparser for the 'predict' command.
        """
        # TODO: modify description later
        desc = """
        Version: todo

        The 'predict' command takes a trained model and input data to run predictions.
        The results are stored in a specified output directory.
        """
        parser = subparsers.add_parser(VNNPredict.COMMAND,
                                       help='Run prediction using a trained model',
                                       description=desc,
                                       formatter_class=constants.ArgParseFormatter)
        parser.add_argument('outdir', help='Directory to write results to')
        parser.add_argument('--inputdir', required=True, help='Path to RO-Crate with the trained model', type=str)
        parser.add_argument('--config_file', help='Config file that can be used to populate arguments for training. '
                                                  'If a given argument is set, it will override the default value.')
        parser.add_argument('--predict_data', help='Path to the dataset to be predicted', type=str)
        parser.add_argument('--gene2id', help='Gene to ID mapping file', type=str)
        parser.add_argument('--cell2id', help='Cell to ID mapping file', type=str)
        parser.add_argument('--mutations', help='Mutation information for cell lines', type=str)
        parser.add_argument('--cn_deletions', help='Copy number deletions for cell lines', type=str)
        parser.add_argument('--cn_amplifications', help='Copy number amplifications for cell lines',
                            type=str)
        parser.add_argument('--batchsize', help='Batchsize', type=int, default=1000)
        parser.add_argument('--zscore_method', help='zscore method (zscore/robustz)', type=str, default='auc')
        parser.add_argument('--cpu_count', help='No of available cores', type=int, default=1)
        parser.add_argument('--drug_count', help='No of top performing drugs', type=int, default=0)
        parser.add_argument('--genotype_hiddens',
                            help='Mapping for the number of neurons in each term in genotype parts', type=int,
                            default=4)
        parser.add_argument('--cuda', help='Specify GPU', type=int, default=0)
        parser.add_argument('--std', help='Standardization File (if not set standardization file from RO-Crate '
                                          'will be used)', type=str)
        parser.add_argument('--slurm', help='If set, slurm script for training will be generated.',
                            action='store_true')
        parser.add_argument('--use_gpu', help='If set, slurm script will be adjusted to run on GPU.',
                            action='store_true')
        parser.add_argument('--slurm_partition', help='Slurm partition. If use_gpu is set, the default is nrnb-gpu.',
                            type=str)
        parser.add_argument('--slurm_account', help='Slurm account. If use_gpu is set, the default is nrnb-gpu.',
                            type=str)
        return parser

    def run(self):
        """
        The logic for running predictions with the model. It executes the prediction process
        using the trained model and input data.

        :raises CellmapsvnnError: If an error occurs during the prediction process.
        """
        try:
            model = os.path.join(self._theargs.inputdir, 'model_final.pt')
            std = os.path.join(self._theargs.inputdir, 'std.txt') if self._theargs.std is None else self._theargs.std
            torch.set_printoptions(precision=5)

            # Load data and model for prediction
            predict_data, cell2id_mapping = self._prepare_predict_data(
                self._theargs.predict_data, self._theargs.cell2id, self._theargs.zscore_method, std)

            # Load cell features
            cell_features = util.load_cell_features(self._theargs.mutations, self._theargs.cn_deletions,
                                                    self._theargs.cn_amplifications)

            hidden_dir = self._get_hidden_dir_path()
            if not os.path.exists(hidden_dir):
                os.mkdir(hidden_dir)

            # Perform prediction
            self.predict(predict_data, model, hidden_dir, self._theargs.batchsize,
                         cell_features)

            hierarchy_file = os.path.join(self._theargs.inputdir, vnnconstants.HIERARCHY_FILENAME)
            factory = RawCX2NetworkFactory()
            hierarchy = factory.get_cx2network(hierarchy_file)
            rlipp_file = os.path.join(self._outdir, vnnconstants.RLIPP_OUTPUT_FILE)
            gene_rho_file = os.path.join(self._outdir, 'gene_rho.out')
            # Perform interpretation
            calc = RLIPPCalculator(hierarchy, self._theargs.predict_data, self._get_predict_dest_file(),
                                   self._theargs.gene2id, self._theargs.cell2id, hidden_dir,
                                   rlipp_file, gene_rho_file, self._theargs.cpu_count, self._theargs.genotype_hiddens,
                                   self._theargs.drug_count, self.excluded_terms)
            calc.calc_scores()
            logger.info('Prediction and interpretation executed successfully')
            print('Prediction and interpretation executed successfully')
        except Exception as e:
            logger.error(f"Error in prediction flow: {e}")
            raise CellmapsvnnError(f"Encountered problem in prediction flow: {e}")

    def _prepare_predict_data(self, test_file, cell2id_mapping_file, zscore_method, std_file):
        """
        Prepares the prediction data for the model.

        :param test_file: Path to the file containing the test dataset.
        :type test_file: str
        :param cell2id_mapping_file: Path to the file containing the cell to ID mapping.
        :type cell2id_mapping_file: str
        :param zscore_method: Method used for z-score standardization.
        :type zscore_method: str
        :param std_file: Path to the standardization file.
        :type std_file: str

        :return: A tuple containing test features and labels as tensors, and the cell2id mapping.
        :rtype: Tuple(Tensor, Tensor), dict
        """
        cell2id_mapping = util.load_mapping(cell2id_mapping_file, 'cell lines')
        test_features, test_labels = self._load_pred_data(test_file, cell2id_mapping, zscore_method, std_file)
        return (torch.Tensor(test_features), torch.Tensor(test_labels)), cell2id_mapping

    @staticmethod
    def _load_pred_data(test_file, cell2id, zscore_method, train_std_file):
        """
        Loads and processes prediction data from a file.

        :param test_file: Path to the file containing the test dataset.
        :type test_file: str
        :param cell2id: Dictionary mapping cell lines to their respective IDs.
        :type cell2id: dict
        :param zscore_method: Method used for z-score standardization.
        :type zscore_method: str
        :param train_std_file: Path to the training standardization file.
        :type train_std_file: str

        :return: Features and labels for the prediction data.
        :rtype: List, List
        """
        train_std_df = pd.read_csv(train_std_file, sep='\t', header=None, names=['dataset', 'center', 'scale'])
        all_df = pd.read_csv(test_file, sep='\t', header=None, names=['cell_line', 'smiles', 'auc', 'dataset'])
        test_df = all_df[all_df['cell_line'].isin(cell2id.keys())]
        test_std_df = util.calc_std_vals(test_df, zscore_method)
        for i, row in test_std_df.iterrows():
            dataset = row['dataset']
            train_entry = train_std_df.query('dataset == @dataset')
            if not train_entry.empty:
                test_std_df.loc[i, 'center'] = float(train_entry['center'].iloc[0])
                test_std_df.loc[i, 'scale'] = float(train_entry['scale'].iloc[0])
        test_df = util.standardize_data(test_df, test_std_df)

        feature = []
        label = []
        for row in test_df.values:
            feature.append([cell2id[row[0]]])
            label.append([float(row[2])])
        return feature, label

    def _get_predict_dest_file(self):
        """
        Returns the file path for saving the prediction results.

        :return: The file path to the prediction results file.
        """
        return os.path.join(self._outdir, 'predict.txt')

    def _get_feature_grad_dest_file(self, grad):
        """
        Returns the file path for saving the gradient of a specific feature.

        :return: The file path to the prediction feature grad file.
        """
        return os.path.join(self._outdir, f'predict_feature_grad_{grad}.txt')

    def _get_hidden_dir_path(self):
        """
        Returns the path to the directory where hidden layer outputs will be stored.

        :return: The file path to the hidden directory.
        """
        return os.path.join(self._outdir, 'hidden/')

    def _to_device(self, tensor):
        if self.use_cuda:
            return tensor.cuda(self._theargs.cuda)
        return tensor

    def predict(self, predict_data, model_file, hidden_folder, batch_size, cell_features=None):
        """
        Perform prediction using the trained model.

        :param predict_data: Tuple of features and labels for prediction.
        :param model_file: Path to the trained model file.
        :param hidden_folder: Directory to store hidden layer outputs.
        :param batch_size: Size of each batch for prediction.
        :param cell_features: Additional cell features for prediction.
        """
        try:
            logger.info('Starting prediction process')
            print('Starting prediction process')
            model = self._load_model(model_file)
            test_loader = self._create_data_loader(predict_data, batch_size)
            test_predict, saved_grads = self._predict(model, test_loader, cell_features, hidden_folder)

            predict_label_gpu = self._to_device(predict_data[1])
            test_corr = util.pearson_corr(test_predict, predict_label_gpu)
            logger.info(f"Test correlation {model.root}: {test_corr:.4f}")

            np.savetxt(self._get_predict_dest_file(), test_predict.cpu().numpy(), '%.4e')

        except Exception as e:
            logger.error(f"Prediction error: {e}")
            raise CellmapsvnnError(f"Encountered problem in prediction: {e}")

    def _load_model(self, model_file):
        """
        Load the trained model for prediction.

        :param model_file: Path to the trained model file.
        :return: Loaded model.
        """
        model = torch.load(model_file,
                           map_location=f'cuda:{self._theargs.cuda}' if self.use_cuda else torch.device("cpu"))
        if self.use_cuda:
            model.cuda(self._theargs.cuda)
        model.eval()
        return model

    def _create_data_loader(self, predict_data, batch_size):
        """
        Create a DataLoader for the prediction data.

        :param predict_data: Tuple of features and labels for prediction.
        :param batch_size: Size of each batch for prediction.
        :return: DataLoader for the prediction data.
        """
        predict_feature, predict_label = predict_data
        return du.DataLoader(du.TensorDataset(predict_feature, predict_label), batch_size=batch_size, shuffle=False)

    def _predict(self, model, data_loader, cell_features, hidden_folder):
        """
        Run the prediction process and save outputs.

        :param model: Trained model for prediction.
        :param data_loader: DataLoader containing the prediction data.
        :param cell_features: Additional cell features for prediction.
        :param hidden_folder: Directory to store hidden layer outputs.
        :return: Tuple of prediction results and saved gradients.
        """
        test_predict = torch.zeros(0, 0)
        if self.use_cuda:
            test_predict = test_predict.cuda(self._theargs.cuda)
        saved_grads = {}

        for i, (inputdata, labels) in enumerate(data_loader):
            cuda_features = self._process_input(inputdata, cell_features)
            aux_out_map, hidden_embeddings_map = model(cuda_features)
            test_predict = torch.cat([test_predict, aux_out_map['final'].data], dim=0) \
                if test_predict.size()[0] else aux_out_map['final'].data

            self._save_hidden_outputs(hidden_embeddings_map, hidden_folder)
            self._register_gradient_hooks(hidden_embeddings_map, saved_grads)
            self._backpropagate(aux_out_map)

            self._save_gradients(cuda_features)
            self._save_hidden_gradients(saved_grads, hidden_folder)

        return test_predict, saved_grads

    def _process_input(self, inputdata, cell_features):
        """
        Process input data for the model.

        :param inputdata: Input data for the model.
        :param cell_features: Additional cell features for prediction.
        :return: Processed features as CUDA variables.
        """
        features = util.build_input_vector(inputdata, cell_features)
        return Variable(self._to_device(features), requires_grad=True)

    def _save_hidden_outputs(self, hidden_embeddings_map, hidden_folder):
        """
        Save outputs from hidden layers.

        :param hidden_embeddings_map: Dictionary of hidden layer outputs.
        :param hidden_folder: Directory to save hidden layer outputs.
        """
        for element, hidden_map in hidden_embeddings_map.items():
            hidden_file = os.path.join(hidden_folder, element + '.hidden')
            with open(hidden_file, 'ab') as f:
                np.savetxt(f, hidden_map.data.cpu().numpy(), '%.4e')

    def _register_gradient_hooks(self, hidden_embeddings_map, saved_grads):
        """
        Register gradient hooks to save gradients of hidden layers.

        :param hidden_embeddings_map: Dictionary of hidden layer outputs.
        :param saved_grads: Dictionary to store saved gradients.
        """

        def save_grad(elem):
            def savegrad_hook(grad):
                saved_grads[elem] = grad

            return savegrad_hook

        for element, _ in hidden_embeddings_map.items():
            hidden_embeddings_map[element].register_hook(save_grad(element))

    def _backpropagate(self, aux_out_map):
        """
        Perform backpropagation.

        :param aux_out_map: Auxiliary output map from the model.
        """
        aux_out_map['final'].backward(torch.ones_like(aux_out_map['final']))

    def _save_gradients(self, cuda_features):
        """
        Save gradients for each feature.

        :param cuda_features: CUDA features variable.
        """
        self._number_feature_grads = len(cuda_features[0, 0, :])
        for i in range(self._number_feature_grads):
            feature_grad = cuda_features.grad.data[:, :, i]
            grad_file = self._get_feature_grad_dest_file(i)
            with open(grad_file, 'ab') as f:
                np.savetxt(f, feature_grad.cpu().numpy(), '%.4e', delimiter='\t')

    def _save_hidden_gradients(self, saved_grads, hidden_folder):
        """
        Save the gradients of the hidden layer outputs.

        :param saved_grads: Dictionary containing the saved gradients.
        :param hidden_folder: Directory to save the hidden layer gradients.
        """
        for element, hidden_grad in saved_grads.items():
            hidden_grad_file = os.path.join(hidden_folder, f'{element}.hidden_grad')
            with open(hidden_grad_file, 'ab') as f:
                np.savetxt(f, hidden_grad.data.cpu().numpy(), '%.4e', delimiter='\t')

    def register_outputs(self, outdir, description, keywords, provenance_utils):
        """
        Registers all output files (predictions, feature gradients, and hidden files)
        with the FAIRSCAPE service for data provenance.

        :param outdir: The directory where the output files are stored.
        :param description: Description for the output files.
        :param keywords: List of keywords associated with the files.
        :param provenance_utils: The utility class for provenance registration.

        :return: A list of dataset IDs for the registered files.
        """
        output_ids = [copy_and_register_gene2id_file(self._theargs.gene2id, outdir, description, keywords,
                                                     provenance_utils),
                      self._register_predict_file(outdir, description, keywords, provenance_utils)]
        for i in range(self._number_feature_grads):
            output_ids.append(self._register_feature_grad_file(outdir, description, keywords, provenance_utils, i))
        output_ids.extend(self._register_hidden_files(outdir, description, keywords, provenance_utils))
        orginal_hierarchy_id = self._copy_and_register_original_hierarchy(outdir, description, keywords, provenance_utils)
        if orginal_hierarchy_id is not None:
            output_ids.append(orginal_hierarchy_id)
        output_ids.append(self._copy_and_register_hierarchy(outdir, description, keywords, provenance_utils))
        id_parent = self._copy_and_register_hierarchy_parent(outdir, description, keywords, provenance_utils)
        if id_parent is not None:
            output_ids.append(id_parent)
        return output_ids

    def _register_predict_file(self, outdir, description, keywords, provenance_utils):
        """
        Registers the prediction result file with the FAIRSCAPE service for data provenance.

        :param outdir: The output directory where the outputs are stored.
        :param description: Description of the file for provenance registration.
        :param keywords: List of keywords associated with the file.
        :param provenance_utils: The utility class for provenance registration.

        :return: The dataset ID assigned to the registered file.
        """
        dest_path = self._get_predict_dest_file()
        description = description
        description += ' prediction result file'
        keywords = keywords
        keywords.extend(['file'])
        data_dict = {'name': os.path.basename(dest_path) + ' prediction result file',
                     'description': description,
                     'keywords': keywords,
                     'data-format': 'txt',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime(provenance_utils.get_default_date_format_str())}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=dest_path,
                                                       data_dict=data_dict)
        return dataset_id

    def _register_feature_grad_file(self, outdir, description, keywords, provenance_utils, grad):
        """
        Registers the feature gradient file with the FAIRSCAPE service for data provenance.

        :param outdir: The output directory where the file is stored.
        :param description: Description of the file for provenance registration.
        :param keywords: List of keywords associated with the file.
        :param provenance_utils: The utility class for provenance registration.
        :param grad: The specific gradient index for the feature.

        :return: The dataset ID assigned to the registered file.
        """
        dest_path = self._get_feature_grad_dest_file(grad)
        description = description
        description += f' prediction feature grad {grad} file'
        keywords = keywords
        keywords.extend(['file'])
        data_dict = {'name': os.path.basename(dest_path) + f' prediction feature grad {grad} file',
                     'description': description,
                     'keywords': keywords,
                     'data-format': 'txt',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime(provenance_utils.get_default_date_format_str())}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=dest_path,
                                                       data_dict=data_dict)
        return dataset_id

    def _register_hidden_files(self, outdir, description, keywords, provenance_utils):
        """
        Registers the output files from the hidden layers with the FAIRSCAPE service for data provenance.

        :param outdir: The directory where the hidden layer output files are stored.
        :param description: A general description for the hidden files.
        :param keywords: A list of keywords associated with the hidden files.
        :param provenance_utils: An instance of the utility class used for handling the provenance registration.

        :return: A list of dataset IDs, each corresponding to a registered hidden file.
        """
        data_dict = {'name': cellmaps_vnn.__name__ + ' hidden layer output file',
                     'description': description + ' hidden layer output file',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime(provenance_utils.get_default_date_format_str())}

        hidden_files_ids = list()

        hidden_dir = self._get_hidden_dir_path()
        cntr = 0
        for entry in tqdm(os.listdir(hidden_dir), desc='FAIRSCAPE hidden files registration'):
            data_dict['data-format'] = entry.split('.')[-1]
            dest_path = os.path.join(hidden_dir, entry)
            data_dict['name'] = os.path.basename(dest_path) + f' hidden file'
            data_dict['keywords'] = ['hidden', 'file']
            dataset_id = provenance_utils.register_dataset(outdir,
                                                           source_file=dest_path,
                                                           data_dict=data_dict)
            hidden_files_ids.append(dataset_id)
            cntr += 1
            if cntr > 5:
                # Todo: https://github.com/fairscape/fairscape-cli/issues/9
                logger.warning('FAIRSCAPE cannot handle too many files, skipping rest')
                break
        return hidden_files_ids

    def _copy_and_register_original_hierarchy(self, outdir, description, keywords, provenance_utils):
        hierarchy_out_file = os.path.join(outdir, vnnconstants.ORIGINAL_HIERARCHY_FILENAME)
        hierarchy_in_file = os.path.join(self._theargs.inputdir, vnnconstants.ORIGINAL_HIERARCHY_FILENAME)
        if not os.path.exists(hierarchy_in_file):
            return None
        shutil.copy(hierarchy_in_file, hierarchy_out_file)

        data_dict = {'name': os.path.basename(hierarchy_out_file) + ' Hierarchy network file',
                     'description': description + ' Hierarchy network file',
                     'keywords': keywords,
                     'data-format': 'CX2',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime('%m-%d-%Y')}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=hierarchy_out_file,
                                                       data_dict=data_dict)
        return dataset_id

    def _copy_and_register_hierarchy(self, outdir, description, keywords, provenance_utils):
        hierarchy_out_file = os.path.join(outdir, vnnconstants.HIERARCHY_FILENAME)
        shutil.copy(os.path.join(self._theargs.inputdir, vnnconstants.HIERARCHY_FILENAME), hierarchy_out_file)

        data_dict = {'name': os.path.basename(hierarchy_out_file) + ' Hierarchy network file used to build VNN',
                     'description': description + ' Hierarchy network file used to build VNN',
                     'keywords': keywords,
                     'data-format': 'CX2',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime('%m-%d-%Y')}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=hierarchy_out_file,
                                                       data_dict=data_dict)
        return dataset_id

    def _copy_and_register_hierarchy_parent(self, outdir, description, keywords, provenance_utils):
        hierarchy_parent_in_file = os.path.join(self._theargs.inputdir, vnnconstants.PARENT_NETWORK_NAME)
        if not os.path.exists(hierarchy_parent_in_file):
            return None
        hierarchy_parent_out_file = os.path.join(outdir, vnnconstants.PARENT_NETWORK_NAME)
        shutil.copy(hierarchy_parent_in_file, hierarchy_parent_out_file)

        data_dict = {'name': os.path.basename(hierarchy_parent_out_file) + ' Hierarchy parent network file',
                     'description': description + ' Hierarchy parent network file',
                     'keywords': keywords,
                     'data-format': 'CX2',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime('%m-%d-%Y')}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=hierarchy_parent_out_file,
                                                       data_dict=data_dict)
        return dataset_id

    def _register_rlipp_file(self, outdir, description, keywords, provenance_utils):
        rlipp_file = os.path.join(outdir, vnnconstants.RLIPP_OUTPUT_FILE)

        data_dict = {'name': os.path.basename(rlipp_file) + ' RLIPP output file',
                     'description': description + ' RLIPP output file',
                     'keywords': keywords,
                     'data-format': 'txt',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime('%m-%d-%Y')}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=rlipp_file,
                                                       data_dict=data_dict)
        return dataset_id

    def _register_gene_rho_file(self, outdir, description, keywords, provenance_utils):
        gene_rho_file = os.path.join(outdir, 'gene_rho.out')

        data_dict = {'name': os.path.basename(gene_rho_file) + ' Gene Rho file',
                     'description': description + ' Gene Rho file',
                     'keywords': keywords,
                     'data-format': 'txt',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime('%m-%d-%Y')}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=gene_rho_file,
                                                       data_dict=data_dict)
        return dataset_id



# train.py
import os
import shutil
from datetime import date

from cellmaps_utils import constants as constants
import cellmaps_vnn.constants as vnnconstants
import logging

import cellmaps_vnn
from cellmaps_vnn.data_wrapper import TrainingDataWrapper
from cellmaps_vnn.exceptions import CellmapsvnnError
from cellmaps_vnn.vnn_trainer import VNNTrainer
from cellmaps_vnn.util import copy_and_register_gene2id_file

logger = logging.getLogger(__name__)


class VNNTrain:
    COMMAND = 'train'

    def __init__(self, theargs):
        """
        Constructor for training Visual Neural Network.
        """
        self._theargs = theargs
        self._outdir = os.path.abspath(theargs.outdir)
        self._theargs.modelfile = self._get_model_dest_file()
        self._theargs.stdfile = self._get_std_dest_file()

    @staticmethod
    def add_subparser(subparsers):
        """
        Adds a subparser for the 'train' command.
        """
        # TODO: modify description later
        desc = """
        Version: todo

        The 'train' command trains a Visual Neural Network using specified hierarchy files
        and data from drugcell or NeSTVNN repository. The resulting model is stored in a
        directory specified by the user.
        """
        parser = subparsers.add_parser(VNNTrain.COMMAND,
                                       help='Train a Visual Neural Network',
                                       description=desc,
                                       formatter_class=constants.ArgParseFormatter)
        parser.add_argument('outdir', help='Directory to write results to')
        parser.add_argument('--inputdir', required=True, help='Path to directory or RO-Crate with hierarchy.cx2 file.'
                                                              'Note that the name of the hierarchy should be '
                                                              'hierarchy.cx2.')
        parser.add_argument('--gene_attribute_name', help='Name of the node attribute of the hierarchy '
                                                          'with list of genes/ proteins of this node. '
                                                          'Default: CD_MemberList.', type=str,
                            default=vnnconstants.GENE_SET_COLUMN_NAME)
        parser.add_argument('--config_file', help='Config file that can be used to populate arguments for training. '
                                                  'If a given argument is set, it will override the default value.')
        parser.add_argument('--training_data', help='Training data')
        parser.add_argument('--gene2id', help='Gene to ID mapping file', type=str)
        parser.add_argument('--cell2id', help='Cell to ID mapping file', type=str)
        parser.add_argument('--mutations', help='Mutation information for cell lines', type=str)
        parser.add_argument('--cn_deletions', help='Copy number deletions for cell lines', type=str)
        parser.add_argument('--cn_amplifications', help='Copy number amplifications for cell lines',
                            type=str)
        parser.add_argument('--batchsize', help='Batchsize', type=int, default=64)
        parser.add_argument('--zscore_method', help='zscore method (zscore/robustz)', type=str, default='auc')
        parser.add_argument('--epoch', help='Training epochs for training', type=int, default=300)
        parser.add_argument('--lr', help='Learning rate', type=float, default=0.001)
        parser.add_argument('--wd', help='Weight decay', type=float, default=0.001)
        parser.add_argument('--alpha', help='Loss parameter alpha', type=float, default=0.3)
        parser.add_argument('--genotype_hiddens',
                            help='Mapping for the number of neurons in each term in genotype parts',
                            type=int, default=4)
        parser.add_argument('--patience', help='Early stopping epoch limit', type=int, default=30)
        parser.add_argument('--delta', help='Minimum change in loss to be considered an improvement', type=float,
                            default=0.001)
        parser.add_argument('--min_dropout_layer', help='Start dropout from this Layer number', type=int, default=2)
        parser.add_argument('--dropout_fraction', help='Dropout Fraction', type=float, default=0.3)
        parser.add_argument('--optimize', help='Hyper-parameter optimization', type=int, default=1)
        parser.add_argument('--cuda', help='Specify GPU', type=int, default=0)
        parser.add_argument('--std', help='Standardization File name', type=str, default='std.txt')
        parser.add_argument('--skip_parent_copy', help='If set, hierarchy parent (interactome) will not be copied',
                            action='store_true')
        parser.add_argument('--slurm', help='If set, slurm script for training will be generated.',
                            action='store_true')
        parser.add_argument('--use_gpu', help='If set, slurm script will be adjusted to run on GPU.',
                            action='store_true')
        parser.add_argument('--slurm_partition', help='Slurm partition. If use_gpu is set, the default is nrnb-gpu.',
                            type=str)
        parser.add_argument('--slurm_account', help='Slurm account. If use_gpu is set, the default is nrnb-gpu.',
                            type=str)
        return parser

    def run(self):
        """
        The logic for training the Visual Neural Network.
        """
        try:
            data_wrapper = TrainingDataWrapper(self._theargs)
            if self._theargs.optimize == 1:
                VNNTrainer(data_wrapper).train_model()
            else:
                logger.error(f"The value {self._theargs.optimize} is wrong value for optimize.")
                raise CellmapsvnnError(f"The value {self._theargs.optimize} is wrong value for optimize.")
        except Exception as e:
            logger.error(f"Training error: {e}")
            raise CellmapsvnnError(f"Encountered problem in training: {e}")

    def _get_model_dest_file(self):
        """
        Returns the file path for saving the trained model file.

        :return: The file path for the model file.
        """
        return os.path.join(self._outdir, 'model_final.pt')

    def _get_std_dest_file(self):
        """
        Returns the file path for saving the standard deviation file.

        :return: The file path for the standard deviation file.
        """
        return os.path.join(self._outdir, self._theargs.std)

    def register_outputs(self, outdir, description, keywords, provenance_utils):
        """
        Registers the model and standard deviation files with the FAIRSCAPE service for data provenance.
        It generates dataset IDs for each registered file.

        :param outdir: The directory where the output files are stored.
        :param description: Description for the output files.
        :param keywords: List of keywords associated with the files.
        :param provenance_utils: The utility class for provenance registration.

        :return: A list of dataset IDs for the registered model and standard deviation files.
        """
        return_ids = [self._register_model_file(outdir, description, keywords, provenance_utils),
                      self._register_std_file(outdir, description, keywords, provenance_utils),
                      self._copy_and_register_hierarchy(outdir, description, keywords, provenance_utils),
                      self._register_pruned_hierarchy(outdir, description, keywords, provenance_utils),
                      copy_and_register_gene2id_file(self._theargs.gene2id, outdir, description, keywords,
                                                     provenance_utils)]
        if not self._theargs.skip_parent_copy:
            id_hierarchy_parent = self._copy_and_register_hierarchy_parent(outdir, description, keywords,
                                                                           provenance_utils)
            if id_hierarchy_parent is not None:
                return_ids.append(id_hierarchy_parent)

        return return_ids

    def _register_model_file(self, outdir, description, keywords, provenance_utils):
        """
        Registers the trained model file with the FAIRSCAPE service for data provenance.

        :param outdir: The output directory where the model file is stored.
        :param description: Description of the model file for provenance registration.
        :param keywords: List of keywords associated with the model file.
        :param provenance_utils: The utility class for provenance registration.

        :return: The dataset ID assigned to the registered model file.
        """
        dest_path = self._get_model_dest_file()
        description = description
        description += ' Model file'
        keywords = keywords
        keywords.extend(['file'])
        data_dict = {'name': os.path.basename(dest_path) + ' trained model file',
                     'description': description,
                     'keywords': keywords,
                     'data-format': 'pt',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime(provenance_utils.get_default_date_format_str())}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=dest_path,
                                                       data_dict=data_dict)
        return dataset_id

    def _register_std_file(self, outdir, description, keywords, provenance_utils):
        """
        Registers the standard deviation file with the FAIRSCAPE service for data provenance.

        :param outdir: The output directory where the standard deviation file is stored.
        :param description: Description of the standard deviation file for provenance registration.
        :param keywords: List of keywords associated with the standard deviation file.
        :param provenance_utils: The utility class for provenance registration.

        :return: The dataset ID assigned to the registered standard deviation file.
        """
        dest_path = self._get_std_dest_file()
        description = description
        description += ' standard deviation file'
        keywords = keywords
        keywords.extend(['file'])
        data_dict = {'name': os.path.basename(dest_path) + ' standard deviation file',
                     'description': description,
                     'keywords': keywords,
                     'data-format': 'txt',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime(provenance_utils.get_default_date_format_str())}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=dest_path,
                                                       data_dict=data_dict)
        return dataset_id

    def _copy_and_register_hierarchy(self, outdir, description, keywords, provenance_utils):
        hierarchy_out_file = os.path.join(outdir, vnnconstants.ORIGINAL_HIERARCHY_FILENAME)
        shutil.copy(os.path.join(self._theargs.inputdir, vnnconstants.HIERARCHY_FILENAME), hierarchy_out_file)

        data_dict = {'name': os.path.basename(hierarchy_out_file) + ' Hierarchy network file',
                     'description': description + ' Hierarchy network file',
                     'keywords': keywords,
                     'data-format': 'CX2',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime('%m-%d-%Y')}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=hierarchy_out_file,
                                                       data_dict=data_dict)
        return dataset_id

    def _register_pruned_hierarchy(self, outdir, description, keywords, provenance_utils):
        hierarchy_out_file = os.path.join(outdir, vnnconstants.HIERARCHY_FILENAME)

        data_dict = {'name': os.path.basename(hierarchy_out_file) + ' Hierarchy network file used to build VNN',
                     'description': description + ' Hierarchy network file used to build VNN',
                     'keywords': keywords,
                     'data-format': 'CX2',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime('%m-%d-%Y')}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=hierarchy_out_file,
                                                       data_dict=data_dict)
        return dataset_id

    def _copy_and_register_hierarchy_parent(self, outdir, description, keywords, provenance_utils):
        hierarchy_parent_in_file = os.path.join(self._theargs.inputdir, vnnconstants.PARENT_NETWORK_NAME)
        if not os.path.exists(hierarchy_parent_in_file):
            logger.warning("No hierarchy parent in the input directory. Cannot copy.")
            return None
        hierarchy_parent_out_file = os.path.join(outdir, vnnconstants.PARENT_NETWORK_NAME)
        shutil.copy(hierarchy_parent_in_file, hierarchy_parent_out_file)

        data_dict = {'name': os.path.basename(hierarchy_parent_out_file) + ' Hierarchy parent network file',
                     'description': description + ' Hierarchy parent network file',
                     'keywords': keywords,
                     'data-format': 'CX2',
                     'author': cellmaps_vnn.__name__,
                     'version': cellmaps_vnn.__version__,
                     'date-published': date.today().strftime('%m-%d-%Y')}
        dataset_id = provenance_utils.register_dataset(outdir,
                                                       source_file=hierarchy_parent_out_file,
                                                       data_dict=data_dict)
        return dataset_id

    @property
    def theargs(self):
        return self._theargs



import unittest
from unittest.mock import MagicMock, patch

import torch

from cellmaps_vnn.data_wrapper import TrainingDataWrapper
from cellmaps_vnn.vnn import VNN
from cellmaps_vnn.vnn_trainer import VNNTrainer


class TestVNNTrainer(unittest.TestCase):
    """Tests for module `vnn_trainer` in `cellmaps_vnn` package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        self.data_wrapper = MagicMock(spec=TrainingDataWrapper)

        self.data_wrapper.train_feature = torch.randn(100, 10)
        self.data_wrapper.train_label = torch.randint(0, 2, (100,))
        self.data_wrapper.val_feature = torch.randn(40, 10)
        self.data_wrapper.val_label = torch.randint(0, 2, (40,))
        self.data_wrapper.cuda = '1'
        self.data_wrapper.batchsize = 16
        self.data_wrapper.epochs = 5
        self.data_wrapper.lr = 0.00
        self.data_wrapper.delta = 0.01
        self.data_wrapper.modeldir = '/path/to/modeldir'
        self.data_wrapper.cell_features = torch.randn(10, 5)
        self.data_wrapper.alpha = 0.5

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_initialization(self):
        vnn_trainer = VNNTrainer(self.data_wrapper)
        self.assertIsNone(vnn_trainer.model)
        self.assertEqual(vnn_trainer.data_wrapper, self.data_wrapper)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `cellmaps_vnn` package."""
import os
import tempfile
import shutil

import unittest
from unittest.mock import MagicMock

from cellmaps_utils.provenance import ProvenanceUtil

from cellmaps_vnn.runner import CellmapsvnnRunner


class TestCellmapsvnnrunner(unittest.TestCase):
    """Tests for `cellmaps_vnn` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_constructor(self):
        """Tests constructor"""
        myobj = CellmapsvnnRunner(outdir='foo', skip_logging=True,
                                  exitcode=0)

        self.assertIsNotNone(myobj)

    def test_run(self):
        """ Tests run()"""
        temp_dir = tempfile.mkdtemp()
        try:
            hier_dir = os.path.join(temp_dir, '4.hierarchy')
            os.makedirs(hier_dir, mode=0o755)
            prov = ProvenanceUtil()
            prov.register_rocrate(hier_dir, name='hierarchy1',
                                  organization_name='hierarchy org',
                                  project_name='hierarchy project')
            cmd = MagicMock()
            cmd.COMMAND = 'train'
            myobj = CellmapsvnnRunner(outdir=os.path.join(temp_dir, 'foo'),
                                      inputdir=hier_dir,
                                      command=cmd,
                                      skip_logging=True,
                                      exitcode=4)
            self.assertEqual(4, myobj.run())
        finally:
            shutil.rmtree(temp_dir)



import unittest
from unittest.mock import Mock, patch

import networkx as nx
import torch

from cellmaps_vnn.data_wrapper import TrainingDataWrapper
from cellmaps_vnn.vnn import VNN


class TestVNN(unittest.TestCase):

    def setUp(self):
        self.mock_data_wrapper = Mock(spec=TrainingDataWrapper)
        self.mock_data_wrapper.root = 'term1'
        self.mock_data_wrapper.num_hiddens_genotype = 10
        self.mock_data_wrapper.gene_id_mapping = {'gene1': 0, 'gene2': 1}
        self.mock_data_wrapper.term_direct_gene_map = {'term1': {'gene1'}, 'term2': {'gene2'}}
        self.mock_data_wrapper.term_size_map = {'term1': 1, 'term2': 1}
        self.mock_data_wrapper.min_dropout_layer = 1
        self.mock_data_wrapper.dropout_fraction = 0.5
        self.mock_data_wrapper.cell_features = torch.randn(100, 2, 10)
        graph = nx.DiGraph()
        graph.add_edge('term1', 'term2')
        self.mock_data_wrapper.digraph = graph

    def test_cal_term_dim(self):
        vnn = VNN(self.mock_data_wrapper)
        expected_term_dim_map = {'term1': 10, 'term2': 10}
        self.assertEqual(vnn.term_dim_map, expected_term_dim_map)

    def test_construct_direct_gene_layer(self):
        vnn = VNN(self.mock_data_wrapper)
        for gene in self.mock_data_wrapper.gene_id_mapping:
            self.assertIn(gene + '_feature_layer', vnn._modules)
            self.assertIn(gene + '_batchnorm_layer', vnn._modules)

    def test_forward(self):
        vnn = VNN(self.mock_data_wrapper)
        input_tensor = torch.randn(100, 2, 10)
        aux_out_map, hidden_embeddings_map = vnn.forward(input_tensor)

        self.assertIsInstance(aux_out_map, dict)
        self.assertIsInstance(hidden_embeddings_map, dict)
        self.assertIn('term1', hidden_embeddings_map)
        self.assertIn('term2', hidden_embeddings_map)

    def test_construct_nn_graph(self):
        vnn = VNN(self.mock_data_wrapper)
        self.assertIsNotNone(vnn.term_layer_list)
        self.assertIsNotNone(vnn.term_neighbor_map)

    def test_term_layer_addition(self):
        vnn = VNN(self.mock_data_wrapper)
        for term in self.mock_data_wrapper.digraph.nodes():
            self.assertIn(term + '_linear_layer', vnn._modules)

    def test_dropout_layer_addition(self):
        vnn = VNN(self.mock_data_wrapper)
        for i, layer in enumerate(vnn.term_layer_list):
            for term in layer:
                if i >= vnn.min_dropout_layer:
                    self.assertIn(term + '_dropout_layer', vnn._modules)

    def test_batchnorm_layer_addition(self):
        vnn = VNN(self.mock_data_wrapper)
        for term in self.mock_data_wrapper.digraph.nodes():
            self.assertIn(term + '_batchnorm_layer', vnn._modules)

    def test_aux_layer_addition(self):
        vnn = VNN(self.mock_data_wrapper)
        for term in self.mock_data_wrapper.digraph.nodes():
            self.assertIn(term + '_aux_linear_layer1', vnn._modules)
            self.assertIn(term + '_aux_linear_layer2', vnn._modules)

    def test_final_layer_addition(self):
        vnn = VNN(self.mock_data_wrapper)
        self.assertIn('final_aux_linear_layer', vnn._modules)
        self.assertIn('final_linear_layer_output', vnn._modules)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import MagicMock, patch

import pandas as pd

from cellmaps_vnn.predict import VNNPredict


class TestPredict(unittest.TestCase):
    """Tests for `predict` in `cellmaps_vnn` package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        self.mock_args = MagicMock()
        self.mock_args.modeldir = '/path/to/model'
        self.mock_args.outdir = '/path/to/output'
        self.mock_args.predict_data = '/path/to/predict_data'
        self.mock_args.cell2id = '/path/to/cell2id'
        self.mock_args.mutations = '/path/to/mutations'
        self.mock_args.cn_deletions = '/path/to/cn_deletions'
        self.mock_args.cn_amplifications = '/path/to/cn_amplifications'
        self.mock_args.batchsize = 1000
        self.mock_args.cuda = 0
        self.mock_args.zscore_method = 'auc'
        self.mock_args.std = None

    def tearDown(self):
        """Tear down test fixtures, if any."""

    @patch('cellmaps_vnn.util.load_mapping')
    @patch('cellmaps_vnn.predict.VNNPredict._load_pred_data')
    def test_prepare_predict_data(self, mock_load_pred_data, mock_load_mapping):
        vnn_predict = VNNPredict(self.mock_args)
        mock_load_mapping.return_value = {'cell_line': 'cell_id'}
        mock_load_pred_data.return_value = ([0.5], [1])
        predict_data, cell2id_mapping = vnn_predict._prepare_predict_data(
            self.mock_args.predict_data, self.mock_args.cell2id, self.mock_args.zscore_method, self.mock_args.std)

        self.assertIsInstance(predict_data, tuple)
        self.assertIsInstance(cell2id_mapping, dict)

    @patch('cellmaps_vnn.predict.pd.read_csv')
    @patch('cellmaps_vnn.util.calc_std_vals')
    @patch('cellmaps_vnn.util.standardize_data')
    def test_load_pred_data(self, mock_standardize_data, mock_calc_std_vals, mock_read_csv):
        mock_train_std_df = pd.DataFrame({
            'dataset': ['dataset1'],
            'center': [0.5],
            'scale': [1.0]
        })
        mock_test_df = pd.DataFrame({
            'cell_line': ['cell1', 'cell2'],
            'smiles': ['smile1', 'smile2'],
            'auc': [0.6, 0.7],
            'dataset': ['dataset1', 'dataset1']
        })
        mock_read_csv.side_effect = [mock_train_std_df, mock_test_df]
        mock_calc_std_vals.return_value = mock_test_df.copy()
        mock_standardize_data.return_value = mock_test_df.copy()
        cell2id = {'cell1': 1, 'cell2': 2}

        feature, label = VNNPredict._load_pred_data('test.csv', cell2id, 'auc', 'train_std.csv')

        self.assertEqual(feature, [[1], [2]])
        self.assertEqual(label, [[0.6], [0.7]])




import time
import logging
import torch.utils.data as du
from torch.autograd import Variable

from cellmaps_vnn.data_wrapper import *
from cellmaps_vnn.vnn import *
from cellmaps_vnn.ccc_loss import *

logger = logging.getLogger(__name__)


class VNNTrainer:
    TRAINING_PROGRESS_FILE = 'training_progress.tsv'

    def __init__(self, data_wrapper):
        """
        Initialize the VNN Trainer.

        :param data_wrapper: data wrapper containing data necessary for training
        :type data_wrapper: TrainingDataWrapper
        """
        self.model = None
        self.data_wrapper = data_wrapper
        self.train_feature = self.data_wrapper.train_feature
        self.train_label = self.data_wrapper.train_label
        self.val_feature = self.data_wrapper.val_feature
        self.val_label = self.data_wrapper.val_label
        self.use_cuda = torch.cuda.is_available() and self.data_wrapper.cuda is not None

    def _to_device(self, tensor):
        """
        Moves a tensor to the appropriate device (GPU or CPU).

        :param tensor: a tensor to be moved.
        :type tensor: torch.Tensor
        :returns: a tensor on GPU or CPU.
        :rtype: torch.Tensor
        """
        if self.use_cuda:
            return tensor.cuda(self.data_wrapper.cuda)
        return tensor

    def train_model(self):
        """
        Trains the VNN model.

        :returns min_loss: The minimum validation loss achieved during training.
        :rtype min_loss: float
        """
        self.model = VNN(self.data_wrapper)
        if self.use_cuda:
            self.model.cuda(self.data_wrapper.cuda)

        term_mask_map = util.create_term_mask(
            self.model.term_direct_gene_map, self.model.gene_dim, self.data_wrapper.cuda
        )
        self._initialize_model_parameters(term_mask_map)

        train_loader = du.DataLoader(du.TensorDataset(self.train_feature, self.train_label),
                                     batch_size=self.data_wrapper.batchsize, shuffle=True, drop_last=True)
        val_loader = du.DataLoader(du.TensorDataset(self.val_feature, self.val_label),
                                   batch_size=self.data_wrapper.batchsize, shuffle=True)

        optimizer = self._configure_optimizer()

        with open(os.path.join(self.data_wrapper.outdir, VNNTrainer.TRAINING_PROGRESS_FILE), "w") as f:
            f.write("epoch\ttrain_corr\ttrain_loss\ttrue_auc\tpred_auc\tval_corr\tval_loss\tgrad_norm\telapsed_time\n")
            min_loss = None

            for epoch in range(self.data_wrapper.epochs):
                epoch_start_time = time.time()
                train_predict, total_loss, gradnorms, train_label_gpu = self._train_epoch(train_loader, optimizer,
                                                                                          term_mask_map)
                val_predict, val_loss, val_label_gpu = self._validate_epoch(val_loader)

                train_corr, val_corr, true_auc, pred_auc = self._calculate_metrics(train_predict, train_label_gpu,
                                                                                   val_predict, val_label_gpu)
                epoch_end_time = time.time()

                elapsed_time = epoch_end_time - epoch_start_time
                f.write(f"{epoch}\t{train_corr:.4f}\t{total_loss:.4f}\t{true_auc:.4f}\t{pred_auc:.4f}\t"
                            f"{val_corr:.4f}\t{val_loss:.4f}\t{gradnorms:.4f}\t{elapsed_time:.4f}\n")

                min_loss = self._save_model_if_improved(min_loss, val_loss, epoch)

        return min_loss

    def _train_epoch(self, train_loader, optimizer, term_mask_map):
        """
        Conducts a training epoch over the training dataset.

        :param train_loader: Data loader for training data.
        :type train_loader: DataLoader
        :param optimizer: Optimizer for the model.
        :type optimizer: Optimizer
        :param term_mask_map: Mask map for term weights.
        :type term_mask_map: Tensor

        :returns : Tuple containing training predictions, total loss, and gradient norms.
        :rtype : Tuple[Tensor, float, float]
        """
        self.model.train()

        train_predict = torch.zeros(0, 0)
        if self.use_cuda:
            train_predict = train_predict.cuda(self.data_wrapper.cuda)
        # tensor for accumulating grad norms from each batch in this epoch
        _gradnorms = torch.empty(len(train_loader))
        if self.use_cuda:
            _gradnorms = _gradnorms.cuda(self.data_wrapper.cuda)
        total_loss = 0
        train_label_gpu = None

        for i, (inputdata, labels) in enumerate(train_loader):
            features, cuda_labels = self._prepare_batch_data(inputdata, labels)
            optimizer.zero_grad()
            aux_out_map, _ = self.model(features)

            if train_predict.size()[0] == 0:
                train_predict = aux_out_map['final'].data
                train_label_gpu = cuda_labels
            else:
                train_predict = torch.cat([train_predict, aux_out_map['final'].data], dim=0)
                train_label_gpu = torch.cat([train_label_gpu, cuda_labels], dim=0)

            total_loss = 0
            for name, output in aux_out_map.items():
                loss = CCCLoss()
                if name == 'final':
                    total_loss += loss(output, cuda_labels)
                else:
                    total_loss += self.data_wrapper.alpha * loss(output, cuda_labels)
            total_loss.backward()

            for name, param in self.model.named_parameters():
                if '_direct_gene_layer.weight' not in name:
                    continue
                term_name = name.split('_')[0]
                param.grad.data = torch.mul(param.grad.data, term_mask_map[term_name])

            # Save gradnorm for batch
            _gradnorms[i] = util.get_grad_norm(self.model.parameters(), 2.0).unsqueeze(0)
            optimizer.step()

        # Save total gradnorm for epoch
        gradnorms = sum(_gradnorms).unsqueeze(0).cpu().numpy()[0]

        return train_predict, total_loss, gradnorms, train_label_gpu

    def _validate_epoch(self, val_loader):
        """
        Conducts a validation epoch over the validation dataset.

        :param val_loader: Data loader for validation data.
        :type val_loader: DataLoader

        :returns : Tuple containing validation predictions and validation loss.
        :rtype : Tuple[Tensor, float]
        """
        self.model.eval()
        val_predict = torch.zeros(0, 0)
        if self.use_cuda:
            val_predict = val_predict.cuda(self.data_wrapper.cuda)
        val_loss = 0
        val_label_gpu = None

        for inputdata, labels in val_loader:
            features, cuda_labels = self._prepare_batch_data(inputdata, labels)
            aux_out_map, _ = self.model(features)

            if val_predict.size()[0] == 0:
                val_predict = aux_out_map['final'].data
                val_label_gpu = cuda_labels
            else:
                val_predict = torch.cat([val_predict, aux_out_map['final'].data], dim=0)
                val_label_gpu = torch.cat([val_label_gpu, cuda_labels], dim=0)

            for name, output in aux_out_map.items():
                loss = CCCLoss()
                if name == 'final':
                    val_loss += loss(output, cuda_labels)

        return val_predict, val_loss, val_label_gpu

    def _prepare_batch_data(self, inputdata, labels):
        """
        Prepares batch data for training or validation.

        :param inputdata: Input data tensor.
        :type inputdata: Tensor
        :param labels: Labels tensor.
        :type labels: Tensor

        :returns : Tuple containing features and labels as Variables.
        :rtype : Tuple[Variable, Variable]
        """
        features = util.build_input_vector(inputdata, self.data_wrapper.cell_features)
        cuda_features = Variable(self._to_device(features))
        cuda_labels = Variable(self._to_device(labels))
        return cuda_features, cuda_labels

    def _initialize_model_parameters(self, term_mask_map):
        """
        Initializes the model parameters, applying term masks and scaling.

        :param term_mask_map: Mask map for term weights.
        :type term_mask_map: Tensor
        """
        for name, param in self.model.named_parameters():
            term_name = name.split('_')[0]
            if '_direct_gene_layer.weight' in name:
                param.data = torch.mul(param.data, term_mask_map[term_name]) * 0.1
            else:
                param.data = param.data * 0.1

    def _configure_optimizer(self):
        """
        Configures the optimizer for the model.

        :returns : Configured optimizer.
        :rtype : torch.optim.Optimizer
        """
        return torch.optim.AdamW(
            self.model.parameters(),
            lr=self.data_wrapper.lr,
            betas=(0.9, 0.99),
            eps=1e-05,
            weight_decay=self.data_wrapper.lr
        )

    @staticmethod
    def _calculate_metrics(train_predict, train_label_gpu, val_predict, val_label_gpu):
        """
        Calculates performance metrics for the training and validation predictions.

        :param train_predict: Training predictions.
        :type train_predict: Tensor
        :param train_label_gpu: Training labels.
        :type train_label_gpu: Tensor
        :param val_predict: Validation predictions.
        :type val_predict: Tensor
        :param val_label_gpu: Validation labels.
        :type val_label_gpu: Tensor

        :returns : Tuple containing training correlation, validation correlation,
                    true average under the curve, predicted average under the curve.
        :rtype : Tuple[float, float, float, float]
        """
        train_corr = util.pearson_corr(train_predict, train_label_gpu)
        val_corr = util.pearson_corr(val_predict, val_label_gpu)
        true_auc = torch.mean(train_label_gpu).item()
        pred_auc = torch.mean(train_predict).item()

        return train_corr, val_corr, true_auc, pred_auc

    def _save_model_if_improved(self, min_loss, val_loss, epoch):
        """
        Saves the model if the validation loss has improved.

        :param min_loss: Minimum loss from previous epochs.
        :type min_loss: float
        :param val_loss: Current epoch's validation loss.
        :type val_loss: float
        :param epoch: Current epoch number.
        :type epoch: int

        :returns : Updated minimum loss.
        :rtype : float
        """
        if min_loss is None or val_loss < min_loss - self.data_wrapper.delta:
            min_loss = val_loss
            torch.save(self.model, self.data_wrapper.modelfile)
            logger.info(f"Model saved at epoch {epoch}")

        return min_loss



import unittest
from unittest.mock import patch, mock_open

from cellmaps_vnn.util import *


class TestUtil(unittest.TestCase):

    def test_calc_std_vals_zscore(self):
        df = pd.DataFrame({'dataset': ['A', 'A', 'B', 'B'],
                           'auc': [1.0, 2.0, 3.0, 4.0]})
        expected_result = pd.DataFrame({
            'dataset': ['A', 'B'],
            'center': [1.5, 3.5],
            'scale': [0.707107, 0.707107]
        }, columns=['dataset', 'center', 'scale'])
        result = calc_std_vals(df, 'zscore')
        pd.testing.assert_frame_equal(result, expected_result)

    def test_calc_std_vals_robustz(self):
        df = pd.DataFrame({'dataset': ['A', 'A', 'B', 'B'],
                           'auc': [1.0, 2.0, 3.0, 4.0]})
        expected_result = pd.DataFrame({
            'dataset': ['A', 'B'],
            'center': [1.5, 3.5],
            'scale': [0.5, 0.5]
        }, columns=['dataset', 'center', 'scale'])
        result = calc_std_vals(df, 'robustz')
        pd.testing.assert_frame_equal(result, expected_result)

    def test_standardize_data(self):
        df = pd.DataFrame({
            'dataset': ['X', 'X', 'Y', 'Y'],
            'auc': [1.0, 2.0, 3.0, 4.0],
            'cell_line': ['CL1', 'CL2', 'CL1', 'CL2'],
            'smiles': ['C1', 'C2', 'C1', 'C2']
        })
        std_df = pd.DataFrame({
            'dataset': ['X', 'Y'],
            'center': [1.5, 3.5],
            'scale': [0.5, 0.5]
        })
        expected_result = pd.DataFrame({
            'cell_line': ['CL1', 'CL2', 'CL1', 'CL2'],
            'smiles': ['C1', 'C2', 'C1', 'C2'],
            'z': [-1.0, 1.0, -1.0, 1.0]
        })
        result = standardize_data(df, std_df)
        pd.testing.assert_frame_equal(result, expected_result)

    @patch('os.path.isfile', return_value=True)
    @patch('numpy.genfromtxt')
    def test_load_numpy_data_success(self, mock_genfromtxt, mock_isfile):
        test_file_path = "test.csv"
        test_data = np.array([[1, 2], [3, 4]])
        mock_genfromtxt.return_value = test_data

        result = load_numpy_data(test_file_path)
        np.testing.assert_array_equal(result, test_data)

    @patch('cellmaps_vnn.util.load_numpy_data')
    def test_load_cell_features(self, mock_load_numpy_data):
        test_mutations = np.array([[1, 0], [0, 1]])
        test_cn_deletions = np.array([[2, 3], [3, 4]])
        test_cn_amplifications = np.array([[4, 5], [5, 6]])

        mock_load_numpy_data.side_effect = [test_mutations, test_cn_deletions, test_cn_amplifications]

        expected_result = np.dstack([test_mutations, test_cn_deletions, test_cn_amplifications])

        result = load_cell_features("mutations.csv", "deletions.csv", "amplifications.csv")
        np.testing.assert_array_equal(result, expected_result)

    def test_build_input_vector(self):
        input_data = torch.tensor([[0], [1]])
        cell_features = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])

        expected_feature = torch.tensor([[[1, 2], [3, 4]], [[5, 6], [7, 8]]], dtype=torch.float)

        result = build_input_vector(input_data, cell_features)
        torch.testing.assert_allclose(result, expected_feature)

    def test_get_grad_norm(self):
        param = torch.tensor([1.0, -2.0, 3.0], requires_grad=True)
        param.grad = torch.tensor([0.1, -0.2, 0.3])

        norm_type = 2
        expected_norm = torch.norm(torch.tensor([0.1, 0.2, 0.3]), norm_type)

        result = get_grad_norm(param, norm_type)
        self.assertAlmostEqual(result.item(), expected_norm.item(), places=5)

    def test_pearson_corr(self):
        x = torch.tensor([1, 2, 3], dtype=torch.float)
        y = torch.tensor([1, 2, 3], dtype=torch.float)

        # Pearson correlation of identical vectors should be 1
        expected_corr = torch.tensor(1.0)

        result = pearson_corr(x, y)
        self.assertAlmostEqual(result.item(), expected_corr.item(), places=5)


if __name__ == '__main__':
    unittest.main()



"""
Contains constants used by cellmaps vnn
"""

PRHO_SCORE = 'P_rho'
"""
P rho score
"""

P_PVAL_SCORE = 'P_pval'
"""
P pval score
"""

CRHO_SCORE = 'C_rho'
"""
C rho score
"""

C_PVAL_SCORE = 'C_pval'
"""
C pval score
"""

RLIPP_SCORE = 'RLIPP'
"""
RLIPP score
"""

RLIPP_OUTPUT_FILE = 'rlipp.out'
"""
Output file from rlipp algorithm
"""

HIERARCHY_FILENAME = 'hierarchy.cx2'
"""
Hierarchy filename.
"""

ORIGINAL_HIERARCHY_FILENAME = 'original_hierarchy.cx2'
"""
Original hierarchy filename.
"""

PARENT_NETWORK_NAME = 'hierarchy_parent.cx2'
"""
Parent network of hierarchy filename.
"""

GENE_SET_COLUMN_NAME = 'CD_MemberList'
"""
Name of the node attribute of the hierarchy with list of genes/ proteins of this node.
"""



import logging
import copy
import torch
import torch.nn as nn
from cellmaps_vnn.data_wrapper import TrainingDataWrapper
from cellmaps_vnn.exceptions import CellmapsvnnError

logger = logging.getLogger(__name__)


class VNN(nn.Module):

    def __init__(self, data_wrapper: TrainingDataWrapper):
        """
        Initializes the VNN model with the provided data wrapper.

        This constructor sets up components of the VNN model, including term maps, gene mappings, dropout
        parameters, and initializes neural network layers based on the given data structure. It also calculates
        the dimensions for each term and constructs the direct gene layers and the neural network graph.

        :param data_wrapper: The necessary data and configurations for initializing the VNN model.
        :type data_wrapper: TrainingDataWrapper
        :raises CellmapsvnnError: If an error occurs during the initialization of the neural network.
        """
        super().__init__()

        self.term_neighbor_map = None
        self.term_layer_list = None
        self.term_dim_map = None
        self.root = data_wrapper.root
        self.num_hiddens_genotype = data_wrapper.num_hiddens_genotype
        self.gene_id_mapping = data_wrapper.gene_id_mapping

        # dictionary from terms to genes directly annotated with the term
        self.term_direct_gene_map = data_wrapper.term_direct_gene_map

        # Dropout Params
        self.min_dropout_layer = data_wrapper.min_dropout_layer
        self.dropout_fraction = data_wrapper.dropout_fraction

        try:
            # calculate the number of values in a state (term): term_size_map is the number of all genes annotated with
            # the term
            self.cal_term_dim(data_wrapper.term_size_map)

            # ngenes, gene_dim are the number of all genes
            self.gene_dim = len(self.gene_id_mapping)

            # No of input features per gene
            self.feature_dim = len(data_wrapper.cell_features[0, 0, :])

            # add modules for neural networks to process genotypes
            self.construct_direct_gene_layer()
            self.construct_nn_graph(copy.deepcopy(data_wrapper.digraph))

            # add module for final layer
            self.add_module('final_aux_linear_layer', nn.Linear(data_wrapper.num_hiddens_genotype, 1))
            self.add_module('final_linear_layer_output', nn.Linear(1, 1))
        except Exception as e:
            raise CellmapsvnnError(f"Error in VNN initialization: {e}")

    def cal_term_dim(self, term_size_map):
        """
        Calculates the dimensionality of each term based on the term sizes.

        This method updates the `term_dim_map` attribute, which maps each term to its dimensionality.
        The dimensionality for each term is set to the number of hidden genotype variables.

        :param term_size_map: A mapping of terms to their sizes.
        :type term_size_map: dict
        """
        self.term_dim_map = {term: int(self.num_hiddens_genotype) for term in term_size_map}

    def construct_direct_gene_layer(self):
        """
        Constructs layers for genes directly annotated with each term.

        This method iterates through each gene and term to create specific layers in the neural network. For each gene,
        it adds a feature layer and a batch normalization layer. For each term, if there are genes directly annotated
        with it, it adds a linear layer that takes all genes as input and outputs only those genes directly annotated
        with the term. If a term has no directly associated genes, the method will raise exception.
        """
        for gene, _ in self.gene_id_mapping.items():
            self.add_module(gene + '_feature_layer', nn.Linear(self.feature_dim, 1))
            self.add_module(gene + '_batchnorm_layer', nn.BatchNorm1d(1))

        for term, gene_set in self.term_direct_gene_map.items():
            if len(gene_set) == 0:
                raise CellmapsvnnError(f'There are no directly associated genes for term: {term}')
            self.add_module(term + '_direct_gene_layer', nn.Linear(self.gene_dim, len(gene_set)))

    def construct_nn_graph(self, digraph):
        """
        Constructs a neural network graph based on given hierarchy.

        This method builds the neural network by starting from the bottom (leaves) of the given directed graph (digraph)
        and iteratively adding modules for each term in the hierarchy. The method stores the built neural network layers
        in `term_layer_list` and maintains a map (`term_neighbor_map`) of each term to its children.

        For each term, the method calculates the input size, which is the sum of the dimensions of its children and the
        number of genes directly annotated by the term. It then adds a series of layers (dropout, linear,
        batch normalization, and auxiliary linear layers) for each term.

        The process continues until all nodes (terms) in the digraph have been processed and added to the network.

        :param digraph: A directed graph representing the ontology, where nodes are terms and edges
                                        indicate term relationships.
        :type digraph: networkx.DiGraph
        """
        self.term_layer_list = []
        self.term_neighbor_map = {}

        for term in digraph.nodes():
            self.term_neighbor_map[term] = []
            for child in digraph.neighbors(term):
                self.term_neighbor_map[term].append(child)

        i = 0
        while True:
            leaves = [n for n in digraph.nodes() if digraph.out_degree(n) == 0]
            if len(leaves) == 0:
                break
            self.term_layer_list.append(leaves)

            for term in leaves:
                input_size = 0
                for child in self.term_neighbor_map[term]:
                    input_size += self.term_dim_map[child]
                if term in self.term_direct_gene_map:
                    input_size += len(self.term_direct_gene_map[term])

                term_hidden = self.term_dim_map[term]

                if i >= self.min_dropout_layer:
                    self.add_module(term + '_dropout_layer', nn.Dropout(p=self.dropout_fraction))
                self.add_module(term + '_linear_layer', nn.Linear(input_size, term_hidden))
                self.add_module(term + '_batchnorm_layer', nn.BatchNorm1d(term_hidden))
                self.add_module(term + '_aux_linear_layer1', nn.Linear(term_hidden, 1))
                self.add_module(term + '_aux_linear_layer2', nn.Linear(1, 1))

            i += 1
            digraph.remove_nodes_from(leaves)

    def forward(self, x):
        """
        Defines the forward function of the VNN model.

        This method processes the input through the neural network constructed in the VNN class. It applies
        a series of transformations to the input data, including feature layer operations, batch normalization,
        and tanh activations. The method aggregates outputs from different terms in the network and finally
        produces two dictionaries: one for hidden embeddings and one for auxiliary outputs.

        :param x: Input tensor representing gene data. Each row corresponds to a gene, and columns are features.
        :type x: torch.Tensor

        :returns: A tuple containing two dictionaries:
                  - hidden_embeddings_map: A mapping from terms to their hidden embeddings.
                  - aux_out_map: A mapping from terms to their auxiliary output.
        :rtype: (dict, dict)
        """
        hidden_embeddings_map = {}
        aux_out_map = {}

        feat_out_list = []
        for gene, i in self.gene_id_mapping.items():
            feat_out = torch.tanh(self._modules[gene + '_feature_layer'](x[:, i, :]))
            hidden_embeddings_map[gene] = self._modules[gene + '_batchnorm_layer'](feat_out)
            feat_out_list.append(hidden_embeddings_map[gene])

        gene_input = torch.cat(feat_out_list, dim=1)
        term_gene_out_map = {}
        for term, _ in self.term_direct_gene_map.items():
            term_gene_out_map[term] = self._modules[term + '_direct_gene_layer'](gene_input)

        for i, layer in enumerate(self.term_layer_list):

            for term in layer:

                child_input_list = []
                for child in self.term_neighbor_map[term]:
                    child_input_list.append(hidden_embeddings_map[child])

                if term in self.term_direct_gene_map:
                    child_input_list.append(term_gene_out_map[term])

                child_input = torch.cat(child_input_list, 1)
                if i >= self.min_dropout_layer:
                    dropout_out = self._modules[term + '_dropout_layer'](child_input)
                    term_nn_out = self._modules[term + '_linear_layer'](dropout_out)
                else:
                    term_nn_out = self._modules[term + '_linear_layer'](child_input)
                tanh_out = torch.tanh(term_nn_out)
                hidden_embeddings_map[term] = self._modules[term + '_batchnorm_layer'](tanh_out)
                aux_layer1_out = torch.tanh(self._modules[term + '_aux_linear_layer1'](hidden_embeddings_map[term]))
                aux_out_map[term] = self._modules[term + '_aux_linear_layer2'](aux_layer1_out)

        final_input = hidden_embeddings_map[self.root]
        aux_layer_out = torch.tanh(self._modules['final_aux_linear_layer'](final_input))
        aux_out_map['final'] = self._modules['final_linear_layer_output'](aux_layer_out)

        return aux_out_map, hidden_embeddings_map



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import os
import re
from setuptools import setup, find_packages


with open(os.path.join('cellmaps_vnn', '__init__.py')) as ver_file:
    for line in ver_file:
        if line.startswith('__version__'):
            version=re.sub("'", "", line[line.index("'"):])

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['cellmaps_utils==0.4.0',
                'ndex2>=3.8.0,<4.0.0',
                'optuna',
                'scikit-learn',
                'networkx',
                'pandas',
                'torch',
                'torchvision',
                'torchaudio',
                'scipy',
                'joblib'
                ]

setup_requirements = []

setup(
    author="Christopher Churas",
    author_email='tools@cm4ai.org',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    description="Python Boilerplate contains all the boilerplate you need to create a Python package with command line",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    long_description_content_type = 'text/x-rst',
    include_package_data=True,
    keywords='cellmaps_vnn',
    name='cellmaps_vnn',
    packages=find_packages(include=['cellmaps_vnn']),
    package_dir={'cellmaps_vnn': 'cellmaps_vnn'},
    package_data={'cellmaps_vnn': ['nest_style.cx2']},
    scripts=[ 'cellmaps_vnn/cellmaps_vnncmd.py'],
    setup_requires=setup_requirements,
    url='https://github.com/idekerlab/cellmaps_vnn',
    version=version,
    zip_safe=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `cellmaps_vnn` package."""

import os
import tempfile
import shutil

import unittest
from cellmaps_vnn import cellmaps_vnncmd


class TestCellmaps_vnn(unittest.TestCase):
    """Tests for `cellmaps_vnn` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_parse_arguments(self):
        """Tests parse arguments"""
        res = cellmaps_vnncmd._parse_arguments('hi', ['train', 'outdir',
                                                      '--inputdir', 'foodir',
                                                      '--training_data', 'footrain',
                                                      '--gene2id', 'foo', '--cell2id', 'foo',
                                                      '--mutations', 'foo', '--cn_deletions', 'foo',
                                                      '--cn_amplifications', 'foo'])

        self.assertEqual('outdir', res.outdir)
        self.assertEqual(1, res.verbose)
        self.assertEqual(0, res.exitcode)
        self.assertEqual(None, res.logconf)

        someargs = ['-vv', '--logconf', 'hi', '--exitcode', '3', 'train', 'outdir',
                    '--inputdir', 'foodir',
                    '--training_data', 'footrain',
                    '--gene2id', 'foo', '--cell2id', 'foo',
                    '--mutations', 'foo', '--cn_deletions', 'foo',
                    '--cn_amplifications', 'foo']
        res = cellmaps_vnncmd._parse_arguments('hi', someargs)

        self.assertEqual('outdir', res.outdir)
        self.assertEqual(3, res.verbose)
        self.assertEqual('hi', res.logconf)
        self.assertEqual(3, res.exitcode)

    def test_main(self):
        """Tests main function"""

        temp_dir = tempfile.mkdtemp()
        # try where loading config is successful
        try:
            outdir = os.path.join(temp_dir, 'out')
            res = cellmaps_vnncmd.main(['myprog.py', '--skip_logging', 'train', outdir,
                                        '--inputdir', 'foodir',
                                        '--training_data', 'footrain',
                                        '--gene2id', 'foo', '--cell2id', 'foo',
                                        '--mutations', 'foo', '--cn_deletions', 'foo',
                                        '--cn_amplifications', 'foo'])
            self.assertEqual(res, 2)
        finally:
            shutil.rmtree(temp_dir)



import math
import os
import logging
import shutil
from datetime import date

import cellmaps_vnn
import numpy as np
import pandas as pd
import torch
from torch import inf

from cellmaps_vnn.exceptions import CellmapsvnnError

logger = logging.getLogger(__name__)


def calc_std_vals(df, zscore_method):
    """
    Calculates standard deviation values for a given DataFrame based on the specified z-score method
    ('zscore' and 'robustz').

    :param df: the data to be standardized.
    :type df: pandas.DataFrame
    :param zscore_method: Method to use for standardization ('zscore' or 'robustz').
    :type zscore_method: str

    :returns std_df: DataFrame with standard deviation values for each dataset.
    :rtype std_df: pandas.DataFrame
    """
    std_df = pd.DataFrame(columns=['dataset', 'center', 'scale'])
    std_list = []

    if zscore_method == 'zscore':
        for name, group in df.groupby(['dataset'])['auc']:
            if isinstance(name, tuple) and len(name) == 1:
                name = name[0]
            center = group.mean()
            scale = group.std()
            if math.isnan(scale) or scale == 0.0:
                scale = 1.0
            temp = pd.DataFrame([[name, center, scale]], columns=std_df.columns)
            std_list.append(temp)

    elif zscore_method == 'robustz':
        for name, group in df.groupby(['dataset'])['auc']:
            if isinstance(name, tuple) and len(name) == 1:
                name = name[0]
            center = group.median()
            scale = group.quantile(0.75) - group.quantile(0.25)
            if math.isnan(scale) or scale == 0.0:
                scale = 1.0
            temp = pd.DataFrame([[name, center, scale]], columns=std_df.columns)
            std_list.append(temp)
    else:
        for name, group in df.groupby(['dataset'])['auc']:
            if isinstance(name, tuple) and len(name) == 1:
                name = name[0]
            temp = pd.DataFrame([[name, 0.0, 1.0]], columns=std_df.columns)
            std_list.append(temp)

    std_df = pd.concat(std_list, ignore_index=True)
    return std_df


def standardize_data(df, std_df):
    """
    Standardizes the data based on provided standard deviation values. This function applies z-score standardization
    to the 'auc' column of the DataFrame, using the standard deviation values provided.

    :param df: the data to be standardized.
    :type df: pandas.DataFrame
    :param std_df: the standard deviation values.
    :type std_df: pandas.DataFrame

    :returns merged: DataFrame with the standardized 'z' values.
    :rtype merged: pandas.DataFrame
    """
    merged = pd.merge(df, std_df, how="left", on=['dataset'], sort=False)
    merged['z'] = (merged['auc'] - merged['center']) / merged['scale']
    merged = merged[['cell_line', 'smiles', 'z']]
    return merged


def load_numpy_data(file_path):
    """
    Reads a file at the specified path and attempts to convert it into a NumPy array.
    If the file is not found or any other error occurs, an exception is raised.

    :param file_path: Path to the file to be loaded.
    :type file_path: str

    :returns: Data loaded from the file.
    :rtype: numpy.ndarray

    :raises CellmapsvnnError: If the file is not found or an error occurs during loading.
    """
    if not os.path.isfile(file_path):
        raise CellmapsvnnError(f"File {file_path} not found.")

    try:
        return np.genfromtxt(file_path, delimiter=',')
    except Exception as e:
        raise CellmapsvnnError(f"Error loading data from {file_path}: {e}")


def load_cell_features(mutations, cn_deletions, cn_amplifications):
    """
    Loads and combines cell/drug features from given mutation, CN deletion, and CN amplification files.

    Each feature set is loaded as a NumPy array and then combined into a single array.

    :param mutations: Path to the mutations data file.
    :type mutations: str
    :param cn_deletions: Path to the CN deletions data file.
    :type cn_deletions: str
    :param cn_amplifications: Path to the CN amplifications data file.
    :type cn_amplifications: str

    :returns: Combined cell features.
    :rtype: numpy.ndarray
    """
    mutations = load_numpy_data(mutations)
    cn_deletions = load_numpy_data(cn_deletions)
    cn_amplifications = load_numpy_data(cn_amplifications)
    return np.dstack([mutations, cn_deletions, cn_amplifications])


def load_mapping(mapping_file, mapping_type):
    """
    Loads a mapping from a file and returns it as a dictionary.

    :param mapping_file: Path to the mapping file.
    :type mapping_file: str
    :param mapping_type: Description of the mapping (e.g., 'gene to ID').
    :type mapping_type: str

    :returns mapping: Dictionary containing the mapping from the file.
    :rtype mapping: dict

    :raises CellmapsvnnError: If the mapping file is not found.
    """
    if not os.path.isfile(mapping_file):
        raise CellmapsvnnError(f"Mapping file {mapping_file} not found.")

    mapping = {}
    file_handle = open(mapping_file)
    for line in file_handle:
        line = line.rstrip().split()
        mapping[line[1]] = int(line[0])

    file_handle.close()
    logger.info('Total number of {} = {}'.format(mapping_type, len(mapping)))
    return mapping


def create_term_mask(term_direct_gene_map, gene_dim, cuda_id=None):
    """
    Creates a term mask map for gene sets. This function generates a mask for each term where the mask is
    a matrix with rows equal to the number of relevant gene set and columns equal to the total number of genes.
    Each element is set to 1 if the corresponding gene is one of the relevant genes.

    :param term_direct_gene_map: Mapping of terms to their respective gene sets.
    :type term_direct_gene_map: dict
    :param gene_dim: Total number of genes.
    :type gene_dim: int
    :param cuda_id: CUDA ID for tensor operations.
    :type cuda_id: int

    :returns term_mask_map: Dictionary of term masks.
    :rtype term_mask_map: dict
    """
    term_mask_map = {}
    for term, gene_set in term_direct_gene_map.items():
        mask = torch.zeros(len(gene_set), gene_dim)
        if cuda_id is not None and torch.cuda.is_available():
            mask = mask.cuda(cuda_id)
        for i, gene_id in enumerate(gene_set):
            mask[i, gene_id] = 1
        term_mask_map[term] = mask
    return term_mask_map


def build_input_vector(input_data, cell_features):
    """
    Builds an input vector for model training using cell features.

    :param input_data: Input data containing cell indices.
    :type input_data: Tensor
    :param cell_features: Cell features array.
    :type cell_features: numpy.ndarray

    :returns feature: Input feature tensor for the model.
    :rtype feature: Tensor
    """
    genedim = len(cell_features[0, :])
    featdim = len(cell_features[0, 0, :])
    feature = np.zeros((input_data.size()[0], genedim, featdim))

    for i in range(input_data.size()[0]):
        feature[i] = cell_features[int(input_data[i, 0])]

    feature = torch.from_numpy(feature).float()
    return feature


def get_grad_norm(model_params, norm_type):
    """
    Computes the gradient norm of model parameters.

    The norm is computed over all gradients together, as if they were
    concatenated into a single vector. Gradients are modified in-place.

    :param model_params: Iterable of model parameters or a single Tensor that will have gradients normalized.
    :type model_params: Iterable[Tensor] or Tensor
    :param norm_type: Type of the p-norm to use (can be 'inf' for infinity norm).
    :type norm_type: float or int

    :returns: Total norm of the model parameters (viewed as a single vector).
    :rtype: Tensor
    """
    if isinstance(model_params, torch.Tensor):  # check if parameters are tensorobject
        model_params = [model_params]  # change to list
    model_params = [p for p in model_params if p.grad is not None]  # get list of params with grads
    norm_type = float(norm_type)  # make sure norm_type is of type float
    if len(model_params) == 0:  # if no params provided, return tensor of 0
        return torch.tensor(0.)

    device = model_params[0].grad.device if torch.cuda.is_available() else torch.device("cpu")  # get device
    if norm_type == inf:  # infinity norm
        total_norm = max(p.grad.detach().abs().max().to(device) for p in model_params)
    else:  # total norm
        total_norm = torch.norm(torch.stack([torch.norm(p.grad.detach(), norm_type).to(device) for p in model_params]),
                                norm_type)
    return total_norm


def pearson_corr(x, y):
    """
    Computes the Pearson correlation coefficient between two tensors.

    :param x: First variable tensor.
    :type x: Tensor
    :param y: Second variable tensor.
    :type y: Tensor

    :returns: Pearson correlation coefficient.
    :rtype: Tensor
    """
    xx = x - torch.mean(x)
    yy = y - torch.mean(y)

    return torch.sum(xx * yy) / (torch.norm(xx, 2) * torch.norm(yy, 2))


def copy_and_register_gene2id_file(genet2id_in_file, outdir, description, keywords, provenance_utils):
    gene2id_out_file = os.path.join(outdir, 'gene2ind.txt')
    shutil.copy(genet2id_in_file, gene2id_out_file)

    data_dict = {'name': os.path.basename(gene2id_out_file) + ' gene to index mapping file',
                 'description': description + ' gene to index mapping file',
                 'keywords': keywords,
                 'data-format': 'txt',
                 'author': cellmaps_vnn.__name__,
                 'version': cellmaps_vnn.__version__,
                 'date-published': date.today().strftime('%m-%d-%Y')}
    dataset_id = provenance_utils.register_dataset(outdir,
                                                   source_file=gene2id_out_file,
                                                   data_dict=data_dict)
    return dataset_id



import os
import shutil
import tempfile
import unittest
import types
from unittest import skip
from unittest.mock import MagicMock, patch

import ndex2.cx2
import networkx as nx
import pandas as pd
import torch

from cellmaps_vnn.data_wrapper import TrainingDataWrapper
import cellmaps_vnn.util as util


class MockArgs:
    def __init__(self, examples_dir, temp_dir):
        self.genotype_hiddens = 128
        self.lr = 0.001
        self.wd = 0.01
        self.alpha = 0.5
        self.epoch = 10
        self.batchsize = 32
        self.outdir = temp_dir
        self.cuda = False
        self.zscore_method = 'standard'
        self.stdfile = os.path.join(temp_dir, 'std.txt')
        self.modelfile = os.path.join(temp_dir, 'model_final.pt')
        self.patience = 5
        self.delta = 0.1
        self.min_dropout_layer = 1
        self.dropout_fraction = 0.3
        self.inputdir = examples_dir
        self.hierarchy = os.path.join(examples_dir, 'hierarchy.cx2')
        self.hierarchy_parent = os.path.join(examples_dir, 'hierarchy_parent.cx2')
        self.training_data = os.path.join(examples_dir, 'training_data.txt')
        self.cell2id = os.path.join(examples_dir, 'cell2ind.txt')
        self.gene2id = os.path.join(examples_dir, 'gene2ind.txt')
        self.mutations = os.path.join(examples_dir, 'cell2mutation.txt')
        self.cn_deletions = os.path.join(examples_dir, 'cell2cndeletion.txt')
        self.cn_amplifications = os.path.join(examples_dir, 'cell2cnamplification.txt')
        self.gene_attribute_name = 'CD_MemberList'


class TestTrainingDataWrapper(unittest.TestCase):

    def setUp(self):
        self.current_dir = os.path.dirname(os.path.abspath(__file__))
        self.project_dir = os.path.dirname(self.current_dir)
        self.examples_dir = os.path.join(self.project_dir, 'examples')
        self.temp_dir = tempfile.mkdtemp()
        self.theargs = MockArgs(self.examples_dir, self.temp_dir)

    def tearDown(self):
        """Tear down test fixtures, if any."""
        shutil.rmtree(self.temp_dir)
        patch.stopall()

    @patch('cellmaps_vnn.data_wrapper.TrainingDataWrapper._prepare_train_data',
           return_value=(['mock_train_feature'], ['mock_train_label'], ['mock_val_feature'], ['mock_val_label']))
    @patch('cellmaps_vnn.data_wrapper.TrainingDataWrapper._load_graph')
    def test_init_with_mocks(self, mock_load_graph, mock_prepare_train_data):
        mock_load_mapping = MagicMock()
        mock_load_numpy_data = MagicMock()
        mock_load_mapping.return_value = "mocked_mapping_value"
        mock_load_numpy_data.return_value = "mocked_numpy_data"
        util.load_mapping = mock_load_mapping
        util.load_numpy_data = mock_load_numpy_data
        wrapper = TrainingDataWrapper(self.theargs)
        self.assertEqual(wrapper._hierarchy, self.theargs.hierarchy)
        self.assertEqual(wrapper.num_hiddens_genotype, self.theargs.genotype_hiddens)
        self.assertEqual(wrapper.lr, self.theargs.lr)
        mock_prepare_train_data.assert_called_once()
        mock_load_graph.assert_called_once()

    def test_init(self):
        wrapper = TrainingDataWrapper(self.theargs)
        self.assertEqual(wrapper._hierarchy, self.theargs.hierarchy)
        self.assertEqual(wrapper.num_hiddens_genotype, self.theargs.genotype_hiddens)
        self.assertEqual(wrapper.lr, self.theargs.lr)
        self.assertIsNotNone(wrapper.train_feature)

    def test_prepare_train_data(self):
        mock_train_features = [[1, 2, 3], [4, 5, 6]]
        mock_train_labels = [0, 1]
        mock_val_features = [[7, 8, 9], [10, 11, 12]]
        mock_val_labels = [1, 0]
        wrapper = MagicMock()
        wrapper._prepare_train_data = types.MethodType(TrainingDataWrapper._prepare_train_data, wrapper)
        wrapper._load_train_data.return_value = (mock_train_features, mock_train_labels, mock_val_features,
                                                 mock_val_labels)
        train_features, train_labels, val_features, val_labels = wrapper._prepare_train_data()
        self.assertTrue(isinstance(train_features, torch.Tensor))
        self.assertTrue(isinstance(train_labels, torch.FloatTensor))
        self.assertTrue(isinstance(val_features, torch.Tensor))
        self.assertTrue(isinstance(val_labels, torch.FloatTensor))

    def test_select_validation_cell_lines(self):
        mock_train_cell_lines = ['CellLine1', 'CellLine2', 'CellLine3', 'CellLine4', 'CellLine5']
        expected_val_size = int(len(mock_train_cell_lines) / 5)
        expected_remaining_train_size = len(mock_train_cell_lines) - expected_val_size
        val_cell_lines = TrainingDataWrapper._select_validation_cell_lines(mock_train_cell_lines)
        self.assertEqual(len(val_cell_lines), expected_val_size)
        for val_cell_line in val_cell_lines:
            self.assertNotIn(val_cell_line, mock_train_cell_lines)
        self.assertEqual(len(mock_train_cell_lines), expected_remaining_train_size)

    def test_create_digraph(self):
        digraph, cx2network = TrainingDataWrapper._create_digraph(self.theargs.hierarchy)
        self.assertTrue(isinstance(digraph, nx.DiGraph))
        self.assertTrue(isinstance(cx2network, ndex2.cx2.CX2Network))


if __name__ == '__main__':
    unittest.main()



import torch


class CCCLoss(torch.nn.Module):
    """
    A PyTorch module for calculating the Concordance Correlation Coefficient (CCC) Loss.

    The CCC Loss is a measure used in regression tasks to evaluate the agreement
    between two variables.
    """

    def __init__(self, eps=1e-6):
        """
        Initializes the CCCLoss module.

        :param eps: A small epsilon value for numerical stability. Default is 1e-6.
        :type eps: float
        """
        super(CCCLoss, self).__init__()
        self.eps = eps

    def forward(self, y_true, y_hat):
        """
        Computes the CCC loss given true and predicted values.

        :param y_true: The true values.
        :type y_true: Tensor
        :param y_hat: The predicted values.
        :type y_hat: Tensor

        :return ccc: The calculated CCC loss.
        :rtype ccc: Tensor
        """
        y_true_mean = torch.mean(y_true)
        y_hat_mean = torch.mean(y_hat)
        y_true_var = torch.var(y_true)
        y_hat_var = torch.var(y_hat)
        y_true_std = torch.std(y_true)
        y_hat_std = torch.std(y_hat)
        vx = y_true - torch.mean(y_true)
        vy = y_hat - torch.mean(y_hat)
        pcc = torch.sum(vx * vy) / (
                torch.sqrt(torch.sum(vx ** 2) + self.eps) * torch.sqrt(torch.sum(vy ** 2) + self.eps))
        ccc = (2 * pcc * y_true_std * y_hat_std) / (y_true_var + y_hat_var + (y_hat_mean - y_true_mean) ** 2)
        ccc = 1 - ccc
        return ccc
